/*
 *	FATX filesystem support (Xbox 360)
 *
 *  Copyright (C) 2012-2024 Christophe Duverger
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "fatx.hpp"

const char * const							sepdir		= "/";				// using unix directories
const char * const							fsid		= "XTAF";			// filesystem id
const char * const							fidx		= "name.txt";		// file used for label name
const char * const							def_landf	= "lost+found";		// default directory for lost & founds
const char * const							def_fpre	= "FILE";			// default file prefix for lost & founds
const char * const							def_label	= "XBOX";			// default label name
const char * const							usb_dir		= "Xbox360";		// USB drive directory
const char * const							usb_data	= "Data";			// USB drive data file prefix


const char * const							mutex_buff   = "Buffer:";
const char * const							mutex_data   = "Data:";
const char * const							mutex_entr   = "Entry:";

fatx_context*				fatx_context::	fatxc		= nullptr;

#ifndef NDEBUG
inline void									dbglog(const string &s) {
	console::write((boost::format("# {%08X} %s") % std::this_thread::get_id() % s).str(), true);
}
#endif

clusptr						vareas::		first() const {
	return empty() ? 0 : begin()->start;
}
clusptr						vareas::		last() const {
	return empty() ? 0 : (end() - 1)->stop;
}
size_t						vareas::		nbcls() const {
	size_t res = 0;
	for(const area& i: *this)
		res += i.stop - i.start + 1;
	return res;
}
clusptr						vareas::		at(size_t s) const {
	if(s == 0)
		return last();
	for(const area& i: *this) {
		if(s <= i.stop - i.start + 1)
			return i.start + s - 1;
		else
			s -= i.stop - i.start + 1;
	}
	return 0;
}
vareas::iterator			vareas::		in(size_t s) {
	if(s == 0)
		return end() - 1;
	for(vareas::iterator i = begin(); i != end(); ++i) {
		if(s <= i->stop - i->start + 1)
			return i;
		else
			s -= i->stop - i->start + 1;
	}
	return end();
}
vareas						vareas::		sub(filesize s, filesize o) const {
	vareas res(*this);
	res.erase(remove_if(
		res.begin(), res.end(),
		[s, o] (const area& a) -> bool {
			return a.offset > o + s - 1 || a.offset + a.size - 1 < o;
		}
	), res.end());
	#if !defined NDEBUG && defined DBG_AREAS
		if(res.empty())
			dbglog("NO SUBAREA.\n");
	#endif
	for(area& i: res) {
		filesize ns = i.size;
		filesize no = i.offset;
		if(o > i.offset && o < i.offset + i.size - 1) {
			no			= o;
			i.pointer	+= o - i.offset;
			ns			-= o - i.offset;
			i.start		+= (o - i.offset) >> fatx_context::get()->par.clus_pow;
		}
		if(o + s > i.offset && o + s < i.offset + i.size) {
			ns			-= i.offset + i.size - o - s;
			i.stop		-= (i.offset + i.size - o - s) >> fatx_context::get()->par.clus_pow;
		}
		i.offset = no;
		i.size = ns;
	}
	#if !defined NDEBUG && defined DBG_AREAS
		if(o > nbcls() << fatx_context::get()->par.clus_pow)
			dbglog((boost::format("AREAS: bad parameters, offset:0x%016X size:%d\n") % o % s).str());
		else
			dbglog("Sub" + res.print());
	#endif
	return o > nbcls() << fatx_context::get()->par.clus_pow ? vareas() : res;
}
void						vareas::		add(vareas va) {
	if(va.empty())
		return;
	if(!empty()) {
		if(va.first() == last() + 1) {
			(end() - 1)->size += va.begin()->size;
			(end() - 1)->stop = va.begin()->stop;
			va.erase(va.begin());
		}
		streamptr o = (end() - 1)->offset + (end() - 1)->size;
		for(area& i: va) {
			i.offset = o;
			o += i.size;
		}
	}
	insert(end(), va.begin(), va.end());
}
void						vareas::		add(clusptr c) {
	if(empty())
		return;
	if(c + 1 == first()) {
		begin()->size++;
		begin()->start = c;
		begin()->pointer = clsarithm::cls2ptr(c);
	}
	else {
		insert(begin(), area(0, clsarithm::cls2ptr(c), 1, c, c));
	}
	for(vareas::iterator i = begin() + 1; i != end(); ++i) {
		i->pointer += fatx_context::get()->par.clus_size;
	}
}
#if !defined NDEBUG && defined DBG_AREAS
string						vareas::		print() const {
	string sink;
	sink += (boost::format("Area: %d subarea%s with %d cluster%s [0x%08X-0x%08X]\n")
		% size()
		% (size() > 1 ? "s" : "")
		% nbcls()
		% (nbcls() > 1 ? "s" : "")
		% first()
		% last()
	).str();
	for(const area& i: *this)
		sink += (boost::format(" [0x%08X-0x%08X] off:0x%016X ptr:0x%016X len:%d\n") % i.start % i.stop % i.offset % i.pointer % i.size).str();
	return sink;
}
#endif

							buffer::		buffer(const streamptr o, const streamptr s) : touched(false), offset(o) {
	if(s == 0) {
		#if !defined NDEBUG && defined DBG_BUFFER
			dbglog((boost::format("... buffer: empty allocation try\n")).str());
		#endif
		return;
	}
	size_t siz = min<streamptr>(static_cast<streamptr>(max_buf), s);
	while(true) {
		resize(siz);
		if(size() >= siz)
			break;
		siz /= 2;
		if(siz < fatx_context::get()->par.clus_size) {
			clear();
			#if !defined NDEBUG && defined DBG_BUFFER
				dbglog((boost::format("... buffer: no way to alloc 0x%08X 0x%016X %d\n") % data() % offset % siz).str());
			#endif
			return;
		}
	}
	#if !defined NDEBUG && defined DBG_BUFFER
		dbglog((boost::format(".oO buffer: 0x%08X 0x%016X %d\n") % this % offset % (*this).size()).str());
	#endif
}
							buffer::		~buffer() {
	#if !defined NDEBUG && defined DBG_BUFFER
		dbglog((boost::format("Oo. buffer: 0x%08X 0x%016X %d\n") % this % offset % (*this).size()).str());
	#endif
}
void						buffer::		enlarge(const streamptr s) {
	if(s > static_cast<streamptr>(max_buf))
		return;
	resize(s);
	if(size() < s) {
		clear();
		#if !defined NDEBUG && defined DBG_BUFFER
			dbglog((boost::format("... buffer: no way to alloc 0x%08X 0x%016X %d\n") % data() % offset % s).str());
		#endif
		return;
	}
	#if !defined NDEBUG && defined DBG_BUFFER
		dbglog((boost::format("OOO buffer: 0x%08X 0x%016X %d\n") % this % offset % (*this).size()).str());
	#endif
}
#if !defined NDEBUG && defined DBGBUFDMP
void						buffer::		operator () (size_t p) {
	string res;
	for(
		size_t i = p;
		i <
			#ifdef DBGBUFDMP
				p + min<string::size_type>(string::size_type(DBGBUFDMP), size());
			#else
				p + size();
			#endif
		i++
	) {
		res += (boost::format(" %02X") % static_cast<unsigned int>(static_cast<unsigned char>((*this)[i]))).str();
		if(((i + 1 - p) % DBGCR) == 0) {
			dbglog(res + "\n");
			res.clear();
		}
	}
	dbglog(res + "\n");
}
#endif

void						mymutx::		lock() {
	if(fatx_context::get()->mmi.prog == frontend::fuse && fatx_context::get()->ready) {
		#if !defined NDEBUG && defined DBG_SEM
			int nb = cpt++;
			#ifdef DBGSEM
				if(nam == DBGSEM)
			#endif
			dbglog((boost::format(">>> Lock request [eXclu:%d %s]\n") % nb % nam).str());
		#endif
		upgrade_mutex::lock();
		#if !defined NDEBUG && defined DBG_SEM
			#ifdef DBGSEM
				if(nam == DBGSEM)
			#endif
			dbglog((boost::format("	Lock done [eXclu:%d %s].\n") % nb % nam).str());
		#endif
	}
}
void						mymutx::		unlock() {
	if(fatx_context::get()->mmi.prog == frontend::fuse && fatx_context::get()->ready) {
		#if !defined NDEBUG && defined DBG_SEM
			int nb = --cpt;
			#ifdef DBGSEM
				if(nam == DBGSEM)
			#endif
			dbglog((boost::format("<<< Lock release [eXclu:%d %s]\n") % nb % nam).str());
		#endif
		upgrade_mutex::unlock();
		#if !defined NDEBUG && defined DBG_SEM
			#ifdef DBGSEM
				if(nam == DBGSEM)
			#endif
			dbglog((boost::format("	Lock done [eXclu:%d %s].\n") % nb % nam).str());
		#endif
	}
}
void						mymutx::		lock_shared() {
	if(fatx_context::get()->mmi.prog == frontend::fuse && fatx_context::get()->ready) {
		#if !defined NDEBUG && defined DBG_SEM
			int nb = cpt++;
			#ifdef DBGSEM
				if(nam == DBGSEM)
			#endif
			dbglog((boost::format(">>> Lock request [Share:%d %s]\n") % nb % nam).str());
		#endif
		upgrade_mutex::lock_shared();
		#if !defined NDEBUG && defined DBG_SEM
			#ifdef DBGSEM
				if(nam == DBGSEM)
			#endif
			dbglog((boost::format("	Lock done [Share:%d %s].\n") % nb % nam).str());
		#endif
	}
}
void						mymutx::		unlock_shared() {
	if(fatx_context::get()->mmi.prog == frontend::fuse && fatx_context::get()->ready) {
		#if !defined NDEBUG && defined DBG_SEM
			int nb = --cpt;
			#ifdef DBGSEM
				if(nam == DBGSEM)
			#endif
			dbglog((boost::format("<<< Lock release [Share:%d %s]\n") % nb % nam).str());
		#endif
		upgrade_mutex::unlock_shared();
		#if !defined NDEBUG && defined DBG_SEM
			#ifdef DBGSEM
				if(nam == DBGSEM)
			#endif
			dbglog((boost::format("	Lock done [Share:%d %s].\n") % nb % nam).str());
		#endif
	}
}

void						console::		write(const string &s, bool err) {
	(err ? cerr : cout) << s;
}
pair<bool, bool>			console::		read() {
	char c, d;
	c = d = static_cast<char>(cin.get());
	while(d != '\n')
		d = static_cast<char>(cin.get());
	return ((c == 'y') || (c == 'Y')) ? make_pair(true, true) : ((c == 'n') || (c == 'N')) ? make_pair(true, false) : make_pair(false, false);
}

							fatx_context::	fatx_context(frontend& m): mmi(m), fat(nullptr), root(nullptr), ready(false) {
}
							fatx_context::	~fatx_context() {
	ready = false;
	destroy();
	set(nullptr);
}
int							fatx_context::	setup() {
	int res = 0;
	if((res = dev.setup()))
		return res;
	if((res = par.setup()))
		return res;
	if(mmi.prog == frontend::fsck || mmi.prog == frontend::unrm || (mmi.prog == frontend::fuse && mmi.recover))
		fat = new memmap(par);
	else
		fat = new dskmap(par);
	if(fat == nullptr)
		return ENOMEM;
	#if !defined NDEBUG && defined DBG_INIT
		dbglog("::EOMAP\n");
	#endif
	if(mmi.prog != frontend::mkfs) {
		root = new entry();
		if(root == nullptr) {
			delete fat;
			return ENOMEM;
		}
		#if !defined NDEBUG && defined DBG_INIT
			dbglog("::EOENT\n");
		#endif
	}
	return res;
}
void						fatx_context::	destroy() {
	delete root;
	root = nullptr;
	delete fat;
	fat = nullptr;
}

							frontend::		frontend(int ac, const char* const * const av) :
	readonly(false),		prog(unknown),			force_y(false),			force_n(false),			force_a(false),
	verbose(false),			recover(false),			local(false),			deldate(true),			dellost(true),
	fuse_debug(false),		fuse_foregrd(false),	fuse_singlethr(false),	nofat(false),			cutname(false),
	argc(ac),				argv(av),				progname(av[0]),		dialog(true),			lostfound(def_landf),
	foundfile(def_fpre),	filecount(0),			mount(),				volname(),				fuse_option(),
	unkopt(),				partition("x2"),		table("hd"),			clus_size(0),			uid(getuid()),
	gid(getgid()),
	mask(
		S_IRUSR | S_IWUSR | S_IXUSR |
		S_IRGRP | S_IXGRP |
		S_IROTH | S_IXOTH
	), allyes(true), offset(0), size(0), input(), script(), diffile() {
}
bool						frontend::		getanswer(bool def) {
	bool res = false;
	console::write(def ? " [Y/n] :" : " [y/N] :");
	if(force_n) {
		console::write("n\n");
		res = false;
	}
	else if(force_y) {
		console::write("y\n");
		res = true;
	}
	else if(force_a) {
		console::write(def ? "y\n" : "n\n");
		res = def;
	}
	else {
		pair<bool, bool> r = console::read();
		res = r.first ? r.second : def;
	}
	allyes = allyes && res;
	return res;
}
string						frontend::		name() {
	return
		prog == fuse ? "fusefatx" :
		prog == mkfs ? "mkfs.fatx" :
		prog == fsck ? "fsck.fatx" :
		prog == unrm ? "unrm.fatx" :
		prog == label ? "label.fatx" :
		"fatx"
	;
}
int							frontend::		setup() {
	size_t pos;
	if((pos = progname.rfind(sepdir, progname.size())) != string::npos)
		progname = progname.substr(pos + 1);
	if(progname.compare("fusefatx") == 0)
		prog	= fuse;
	if(progname.compare("mkfs.fatx") == 0) {
		prog	= mkfs;
		dialog	= false;
	}
	if(progname.compare("fsck.fatx") == 0) {
		prog	= fsck;
		dialog	= false;
	}
	if(progname.compare("unrm.fatx") == 0) {
		prog	= unrm;
		recover	= true;
		dialog	= false;
	}
	if(progname.compare("label.fatx") == 0) {
		prog	= label;
		dialog	= false;
	}
	options_description hidden;
	variables_map varmap;
	hidden.add_options()
		("default", "display options default values")
		("as", value<string>(),
			"choose program behavior\n"
			"\"fuse\"  for fusefatx,\n"
			"\"fsck\"  for fsck.fatx,\n"
			"\"mkfs\"  for mkfs.fatx,\n"
			"\"unrm\"  for unrm.fatx,\n"
			"\"label\" for label.fatx"
		)
		("do", value<string>(), "send a script to the program")
	;
	vector<string> visopt;
	try {
		parsed_options parsed(command_line_parser(argc, argv)
			.options(hidden)
			.allow_unregistered()
			.run()
		);
		store(parsed, varmap);
		visopt = collect_unrecognized(parsed.options, include_positional);
		notify(varmap);
	}
	catch(std::exception& e) {
		ostringstream s;
		s << hidden;
		console::write(s.str());
		s.str("");
		s << e.what();
		console::write(s.str() + "\n");
		prog = unknown;
	}
	if(varmap.count("as")) {
		if(varmap["as"].as<string>() == "fuse")
			prog = fuse;
		if(varmap["as"].as<string>() == "mkfs")
			prog = mkfs;
		if(varmap["as"].as<string>() == "fsck")
			prog = fsck;
		if(varmap["as"].as<string>() == "unrm")
			prog = unrm;
		if(varmap["as"].as<string>() == "label")
			prog = label;
	}
	if(varmap.count("do"))
		script = varmap["do"].as<string>();
	if(prog == mkfs || prog == fsck || prog == unrm || prog == label)
		dialog	= false;
	if(prog == unrm)
		recover	= true;
	#ifndef NDEBUG
		dbglog((boost::format("=> CALL: %s\n") % name()).str());
	#endif
	if(prog != fuse && prog != fsck && prog != mkfs && prog != unrm && prog != label) {
		console::write(
			"Invalid usage.\n"
			"Please use a link to this executable named:\n"
			"- fusefatx\tto mount a filesystem with fuse\n"
			"- mkfs.fatx\tto create a new filesystem\n"
			"- fsck.fatx\tto check a filesystem\n"
			"- unrm.fatx\tto try to recover deleted files\n"
			"- label.fatx\tto display or change a filesystem label\n"
		);
		return EINVAL;
	}
	options_description visible((boost::format("Usage: %s [options] device%s")
		% name()
		% (prog == fuse ? " mountpoint" : (prog == label ? " [label]" : ""))
	).str());
	visible.add_options()
		("help,h", "display this help message")
		("version", "display the version number")
		("verbose,v", "verbose output")
		("input,i", value<string>(), "set input device/file")
		("diff", value<string>(), "set separate write file")
		("offset", value<streamptr>(), "force partition offset")
		("size", value<streamptr>(), "force partition size")
		("partition,p", value<string>()->default_value("x2"),
			"select partition:\n"
			"\"sc\"  for system cache partition,\n"
			"\"gc\"  for game cache partition,\n"
			"\"se1\" for sysext partition,\n"
			"\"se2\" for sysext2 partition,\n"
			"\"xdv\" for Xbox 360 dashboard partition,\n"
			"\"x1\"  for original Xbox compatibility partition,\n"
			"\"x2\"  for data partition (default)"
		)
		("table,b", value<string>()->default_value("hd"),
			"select partition table:\n"
			"\"file\" for plain file,\n"
			"\"mu\"   for memory unit,\n"
			"\"usb\"  for USB drive,\n"
			"\"hd\"   for Xbox 360 retail hard disk (default),\n"
			"\"kit\"  for devkit hard disk"
		 )
	;
	if(prog == fuse) {
		visible.add_options()
			("mount,m", value<string>(), "set mountpoint")
			("recover,r", "mount with deleted files")
			("option,o", value<string>(), "mount options")
			("cutname,c", "enable use of long names")
			("debug,d", "enable debug output (implies -f)")
			("foregrd,f", "foreground operation")
			("singlethr,s", "fuse on single thread")
			("uid",  value<uid_t>(), "sets uid of the filesystem")
			("gid",  value<gid_t>(), "sets gid of the filesystem")
			("mask",  value<string>(), "sets mask for entries modes")
		;
	}
	if(prog == label || prog == mkfs) {
		visible.add_options()
			("label,l", value<string>(), "set volume label")
		;
	}
	if(prog == mkfs) {
		visible.add_options()
			("cls-size,c", value<streamptr>(), "set num of blocks per cluster")
		;
	}
	if(prog == fsck || prog == unrm || prog == mkfs) {
		visible.add_options()
			("all,y", "answer yes to everything")
			("none,n", "answer no to everything")
			("auto,a", "default answer to everything")
		;
	}
	if(prog == fsck || prog == unrm || prog == mkfs || prog == fuse) {
		visible.add_options()
			("test,t", "test mode, no modification done")
		;
	}
	if(prog == unrm) {
		visible.add_options()
			("local,l", "recover files in local filesystem")
		;
	}
	if(prog == fsck || prog == unrm) {
		visible.add_options()
			("nofat,f", "disable FAT sanity check and recovery")
		;
	}
	if(prog == fuse || prog == unrm) {
		visible.add_options()
			("nodate", "disable dates precedence of deleted files")
			("nolost", "disable preservation of lost chains")
		;
	}
	positional_options_description	podesc;
	podesc.add("input", 1);
	if(prog == fuse)
		podesc.add("mount", 1);
	if(prog == label)
		podesc.add("label", 1);
	try {
		if(prog == fuse) {
			parsed_options parsed(command_line_parser(visopt)
				.options(visible)
				.positional(podesc)
				.allow_unregistered()
				.run()
			);
			store(parsed, varmap);
			unkopt = collect_unrecognized(parsed.options, exclude_positional);
		}
		else {
			parsed_options parsed(command_line_parser(visopt)
				.options(visible)
				.positional(podesc)
				.run()
			);
			store(parsed, varmap);
		}
		notify(varmap);
	}
	catch(std::exception& e) {
		ostringstream s;
		s << visible;
		console::write(s.str());
		s.str("");
		s << e.what();
		console::write(s.str() + "\n");
		prog = unknown;
	}
	if(varmap.count("all"))
		force_y		= true;
	if(varmap.count("none"))
		force_n		= true;
	if(varmap.count("auto"))
		force_a		= true;
	if(varmap.count("test"))
		readonly	= true;
	if(varmap.count("verbose"))
		verbose		= true;
	if(varmap.count("cutname"))
		cutname		= true;
	if(varmap.count("recover")) {
		recover		= true;
		readonly	= true;
	}
	if(varmap.count("local")) {
		local		= true;
		readonly	= true;
	}
	if(varmap.count("nofat")) {
		nofat		= true;
	}
	if(varmap.count("nodate")) {
		deldate		= false;
		if(prog == fuse) {
			recover		= true;
			readonly	= true;
		}
	}
	if(varmap.count("nolost")) {
		dellost		= false;
		if(prog == fuse) {
			recover		= true;
			readonly	= true;
		}
	}
	if(varmap.count("debug")) {
		fuse_debug		= true;
		fuse_foregrd	= true;
	}
	if(varmap.count("singlethr"))
		fuse_singlethr	= true;
	if(varmap.count("foregrd"))
		fuse_foregrd	= true;
	if(varmap.count("mount"))
		mount			= varmap["mount"].as<string>();
	if(varmap.count("label"))
		volname			= varmap["label"].as<string>();
	if(varmap.count("partition"))
		partition		= varmap["partition"].as<string>();
	if(varmap.count("table"))
		table			= varmap["table"].as<string>();
	if(varmap.count("cls-size"))
		clus_size		= varmap["cls-size"].as<streamptr>();
	if(varmap.count("uid"))
		uid				= varmap["uid"].as<uid_t>();
	if(varmap.count("gid"))
		gid				= varmap["gid"].as<gid_t>();
	if(varmap.count("mask")) {
		stringstream ss;
		ss << std::oct << varmap["mask"].as<string>();
		ss >> mask;
	}
	if(varmap.count("offset"))
		offset			= varmap["offset"].as<streamptr>();
	if(varmap.count("size"))
		size			= varmap["size"].as<streamptr>();
	if(varmap.count("input"))
		input			= varmap["input"].as<string>();
	if(varmap.count("diff"))
		diffile			= varmap["diff"].as<string>();
	if(prog == label)
		readonly		= !varmap.count("label");
	if(varmap.count("option")) {
		tokenizer<char_separator<char> > opts(varmap["option"].as<string>(), char_separator<char>(","));
		for(const string& o: opts) {
			if(o == "ro")
				readonly = true;
			else {
				if(!fuse_option.empty())
					fuse_option += ",";
				fuse_option += o;
			}
		}
	}
	if(varmap.count("version")) {
		console::write((boost::format("%s v%s.\nCopyright (C) 2012 - 2023 Christophe Duverger.\n\n")
			% name()
			% PACKAGE_VERSION
		).str());
		console::write(
			"This program comes with ABSOLUTELY NO WARRANTY.\n"
			"This is free software, and you are welcome to redistribute it\n"
			"under certain conditions.\n"
		);
		prog = unknown;
		return 0;
	}
	if(varmap.count("default")) {
		console::write(
			(boost::format("diffile\t\t%s\n")		% (diffile.empty() ? "none" : diffile)).str() +
			(boost::format("force yes\t%s\n")		% (force_y ? "yes" : "no")).str() +
			(boost::format("force no\t%s\n")		% (force_n ? "yes" : "no")).str() +
			(boost::format("force auto\t%s\n")		% (force_a ? "yes" : "no")).str() +
			(boost::format("verbose\t\t%s\n")		% (verbose ? "yes" : "no")).str() +
			(boost::format("read only\t%s\n")		% (readonly ? "yes" : "no")).str() +
			(boost::format("fuse recover\t%s\n")	% (recover ? "yes" : "no")).str() +
			(boost::format("local recover\t%s\n")	% (local ? "yes" : "no")).str() +
			(boost::format("deleted dates\t%s\n")	% (deldate ? "yes" : "no")).str() +
			(boost::format("preserve losts\t%s\n")	% (dellost ? "yes" : "no")).str() +
			(boost::format("ignore fat\t%s\n")		% (nofat ? "yes" : "no")).str() +
			(boost::format("cut long names\t%s\n")	% (cutname ? "yes" : "no")).str() +
			(boost::format("partition\t%s\n")		% partition).str() +
			(boost::format("table\t\t%s\n")		% (table.empty() ? "file" : table)).str() +
			(boost::format("offset\t\t%d\n")		% offset).str() +
			(boost::format("size\t\t%d\n")			% size).str() +
			(boost::format("cluster size\t%d\n")	% clus_size).str() +
			(boost::format("fuse debug\t%s\n")		% (fuse_debug ? "yes" : "no")).str() +
			(boost::format("fuse foregrd\t%s\n")	% (fuse_foregrd ? "yes" : "no")).str() +
			(boost::format("fuse singlethr\t%s\n")	% (fuse_singlethr ? "yes" : "no")).str() +
			(boost::format("uid\t\t%d\n")			% uid).str() +
			(boost::format("gid\t\t%d\n")			% gid).str() +
			(boost::format("mask\t\t%03o\n")		% mask).str()
		);
		return EPERM;
	}
	if(prog == mkfs && table == "usb") {
		console::write("Can't make filesystem for USB drive.\n");
		return EINVAL;
	}
	if(varmap.count("help") || !varmap.count("input") || (prog == fuse && !varmap.count("mount"))) {
		ostringstream s;
		s << visible;
		console::write(s.str());
		return EINVAL;
	}
	#if !defined NDEBUG && defined DBG_INIT
		dbglog("::EOMMI\n");
	#endif
	return 0;
}
void						frontend::		parser() {
	script.erase(remove_if(script.begin(), script.end(), [] (const char c) ->bool { return c == ' ' || c == '\t' || c == '\n'; }), script.end());
	tokenizer<char_separator<char> > cmds(script, char_separator<char>(";"));
	for (const string &cmd : cmds) {
		tokenizer<char_separator<char> > args(cmd, char_separator<char>(","));
		tokenizer<char_separator<char> >::iterator i = args.begin();
		if(i->empty())
			break;
		else if((*i)[0] == '#')
			continue;
		else if(*i == "mkdir" && ++i != args.end() && !i->empty()) {
			console::write("mkdir:");
			if(!writeable()) {
				console::write("read-only\n");
				continue;
			}
			size_t l = i->find_last_of(sepdir);
			if(l == string::npos || l == i->length() - 1) {
				console::write("syntax error\n");
				continue;
			}
			entry* n = new entry(i->substr(l + 1), 0, true);
			entry* s = fatx_context::get()->root->find(i->substr(0, l).data());
			if(s == nullptr) {
				console::write("not found\n");
				continue;
			}
			s->addtodir(n);
			console::write(n->path() + "\n");
		}
		else if(*i == "rmdir" && ++i != args.end() && !i->empty()) {
			console::write("rmdir:");
			if(!writeable()) {
				console::write("read-only\n");
				continue;
			}
			entry* n = fatx_context::get()->root->find(i->data());
			if(n == nullptr || !n->flags().dir) {
				console::write("not found\n");
				continue;
			}
			if(n->childs.size() != 0) {
				console::write("not empty\n");
				continue;
			}
			console::write(n->path() + "\n");
			n->parent()->remfrdir(n);
		}
		else if(*i == "cp" && ++i != args.end() && !i->empty()) {
			console::write("cp:");
			if(!writeable()) {
				console::write("read-only\n");
				continue;
			}
			entry* s = fatx_context::get()->root->find(i++->data());
			if(s == nullptr) {
				console::write("not found\n");
				continue;
			}
			size_t l = i->find_last_of(sepdir);
			if(l == string::npos || l == i->length() - 1) {
				console::write("syntax error\n");
				continue;
			}
			entry* d = fatx_context::get()->root->find(i->substr(0, l).data());
			if(d == nullptr) {
				console::write("not found\n");
				continue;
			}
			entry* n = new entry(i->substr(l + 1), s->size());
			d->addtodir(n);
			string b(s->size(), '\0');
			s->data(&b[0], true, 0, s->size());
			n->data(&b[0], false, 0, n->size());
			console::write(n->path() + "\n");
		}
		else if(*i == "rcp" && ++i != args.end() && !i->empty()) {
			console::write("rcp:");
			ifstream s(i++->data(), ios::binary);
			if(!s) {
				console::write("can't open\n");
				continue;
			}
			s.seekg(0, ios::end);
			streamptr siz = static_cast<streamptr>(s.tellg());
			s.seekg(0, ios::beg);
			size_t l = i->find_last_of(sepdir);
			if(l == string::npos || l == i->length() - 1) {
				console::write("syntax error\n");
				continue;
			}
			entry* d = fatx_context::get()->root->find(i->substr(0, l).data());
			if(d == nullptr) {
				console::write("not found\n");
				continue;
			}
			entry* n = new entry(i->substr(l + 1), siz);
			d->addtodir(n);
			string b(siz, '\0');
			s.read(&b[0], static_cast<streamsize>(siz));
			s.close();
			console::write((boost::format("(%d)") % siz).str());
			n->data(&b[0], false, 0, n->size());
			console::write(n->path() + "\n");
		}
		else if(*i == "lcp" && ++i != args.end() && !i->empty()) {
			console::write("lcp:");
			entry* s = fatx_context::get()->root->find(i++->data());
			if(s == nullptr) {
				console::write("not found\n");
				continue;
			}
			ifstream t;
			t.open(i->data());
			if(t) {
				t.close();
				console::write("local file exists\n");
				continue;
			}
			ofstream d(i->data(), ios::binary | ios::trunc);
			d.seekp(0, ios::beg);
			string b(s->size(), '\0');
			s->data(&b[0], true, 0, s->size());
			console::write((boost::format("(%d)") % s->size()).str());
			d.write(&b[0], static_cast<streamsize>(s->size()));
			d.close();
			console::write(*i + "\n");
		}
		else if(*i == "mv" && ++i != args.end() && !i->empty()) {
			console::write("mv:");
			if(!writeable()) {
				console::write("read-only\n");
				continue;
			}
			entry* n = fatx_context::get()->root->find(i++->data());
			if(n == nullptr) {
				console::write("not found\n");
				continue;
			}
			n->rename(i->data());
			console::write(n->path() + "\n");
		}
		else if(*i == "rm" && ++i != args.end() && !i->empty()) {
			console::write("rm:");
			if(!writeable()) {
				console::write("read-only\n");
				continue;
			}
			entry* n = fatx_context::get()->root->find(i->data());
			if(n == nullptr || n->flags().dir) {
				console::write("not found\n");
				continue;
			}
			console::write(n->path() + "\n");
			n->parent()->remfrdir(n);
		}
		else if(*i == "lsfat" && ++i != args.end() && !i->empty()) {
			console::write(*i + ":");
			entry* e = fatx_context::get()->root->find(i->data());
			if(e != nullptr)
				console::write(fatx_context::get()->fat->printchain(e->cluster()));
			else
				console::write("not found");
			console::write("\n");
		}
		else if(*i == "mklost") {
			clusptr p = 0;
			console::write(*i + ":");
			if(!writeable()) {
				console::write("read-only\n");
				continue;
			}
			while(++i != args.end()) {
				clusptr s = 0;
				clusptr e = 0;
				size_t l = i->find_last_of(":");
				if(l != string::npos) {
					try {
						s = lexical_cast<clusptr>(i->substr(0, l));
					}
					catch(bad_lexical_cast &) {
						console::write("*ERR*");
						break;
					}
					try {
						e = lexical_cast<clusptr>(i->substr(l + 1));
					}
					catch(bad_lexical_cast &) {
						console::write("*ERR*");
						break;
					}
				}
				else {
					try {
						s = lexical_cast<clusptr>(*i);
					}
					catch(bad_lexical_cast &) {
						console::write("*ERR*");
						break;
					}
				}
				do {
					if(p != 0) {
						console::write((boost::format("0x%08X->") % p).str());
						fatx_context::get()->fat->write(p, s);
					}
					p = s++;
				} while(e != 0 && p != e);
			}
			if(p != 0) {
				console::write((boost::format("0x%08X->EOC") % p).str());
				fatx_context::get()->fat->write(p, EOC);
			}
			console::write("\n");
		}
		else if(*i == "rmfat") {
			console::write(*i + ":");
			if(!writeable()) {
				console::write("read-only\n");
				continue;
			}
			while(++i != args.end()) {
				clusptr s = 0;
				clusptr e = 0;
				size_t l = i->find_last_of(":");
				if(l != string::npos) {
					try {
						s = lexical_cast<clusptr>(i->substr(0, l));
					}
					catch(bad_lexical_cast &) {
						console::write("*ERR*");
						break;
					}
					try {
						e = lexical_cast<clusptr>(i->substr(l + 1));
					}
					catch(bad_lexical_cast &) {
						console::write("*ERR*");
						break;
					}
				}
				else {
					try {
						s = lexical_cast<clusptr>(*i);
					}
					catch(bad_lexical_cast &) {
						console::write("*ERR*");
						break;
					}
				}
				do {
					console::write((boost::format("0x%08X ") % s).str());
					fatx_context::get()->fat->write(s++, FLK);
				} while(e != 0 && s != e);

			}
			console::write("\n");
		}
		else if(*i == "chcls" && ++i != args.end() && !i->empty()) {
			entry* e = fatx_context::get()->root->find(i->data());
			if(e != nullptr) {
				console::write(*i + ":");
				if(!writeable()) {
					console::write("read-only\n");
					continue;
				}
				if(++i != args.end()) {
					clusptr n = 0;
					try {
						n = lexical_cast<clusptr>(*i);
					}
					catch(bad_lexical_cast &) {
						console::write("*ERR*");
						break;
					}
					e->cluster(n);
					e->save();
					console::write((boost::format("%s->0x%08X") % e->path() % e->cluster()).str());
				}
			}
			else
				console::write("not found");
			console::write("\n");
		}
		else if(*i == "help") {
			console::write(
				"syntax: cmd, arg1, arg2, ...[; cmd, arg1, ...[; ...]]\n"
				"\tmkdir,\t/path/to/newdir\n"
				"\trmdir,\t/path/to/dir\n"
				"\tcp,\t/path/to/src, /path/to/dst\n"
				"\trcp,\t/path/to/local/src, /path/to/dst\n"
				"\tlcp,\t/path/to/src, /path/to/local/dst\n"
				"\tmv,\t/path/to/src, /path/to/dst\n"
				"\trm,\t/path/to/file\n"
				"\tlsfat,\t/path/to/file\n"
				"\tmklost,\tclus, start:end, ...\n"
				"\trmfat,\tclus, start:end, ...\n"
				"\tchcls,\t/path/to/file, clus\n"
				"\t#comment, ...\n"
			);
		}
		else
			console::write(*i + ":unknown" + "\n");
	}
}

bool					device::chgfile::	load() {
	streamptr o = 0;
	streamptr p;
	streamptr s;
	while(iod.seekg(static_cast<std::basic_istream<char>::off_type>(o), ios::beg), iod.peek()) {
		if(!iod.read(reinterpret_cast<char*>(&p), sizeof(p)))
			return true;
		if(!iod.read(reinterpret_cast<char*>(&s), sizeof(s)) || s == 0)
			return true;
		device::segment seg = { s, (o += 2 * sizeof(streamptr)) };
		insert(make_pair(p, seg));
		o += s;
	}
	#if !defined NDEBUG && defined DBG_INIT
		dbglog((boost::format("Diff file loaded %s\n") % (size() == 0 ? "(0 segment)." : (boost::format("(%d segment%s):") % size() % (size() == 1 ? "" : "s")).str())).str());
		for(auto i: *this)
			dbglog((boost::format("0x%016X (%d) at 0x%016X\n") % i.first % i.second.size % i.second.position).str());
	#endif
	return false;
}
bool					device::chgfile::	read(streamptr p, const size_t s, string &buf) {
	#if !defined NDEBUG && defined DBG_DIFF
		dbglog((boost::format("=>  Read diffile at 0x%016X(%d)\n") % p % s).str());
	#endif
	auto b = lower_bound(p);		// *b >= p
	if(b != begin()) {
		b--;						// *b < p && *b + b->s > p
		if(b->first + b->second.size <= p)
			b++;
	}
	auto e = lower_bound(p + s);	// *e > p + s - 1
	#if !defined NDEBUG && defined DBG_DIFF
		dbglog(
			"	match: " +
			(b == end() ? "end" : (boost::format("0x%016X[0x%016X(%d)]") % b->second.position % b->first % b->second.size).str()) +
			" - " +
			(e == end() ? "end" : (boost::format("0x%016X[0x%016X(%d)]") % e->second.position % e->first % e->second.size).str()) +
			"\n"
		);
	#endif
	streamptr c = 0;
	for(auto i = b; i != e; i++) {
		auto pos = i->second.position;
		if(i->first < p)
			pos += p - i->first;
		if(i->first > p + c)
			c = i->first - p;
		auto siz = min<filesize>(buf.size() - c, i->second.size - (i->first < p ? p - i->first : 0));
		#if !defined NDEBUG && defined DBG_DIFF
			dbglog((boost::format("... read diffile at 0x%016X [0x%016X(%d)]\n") % pos % (p + c) % siz).str());
		#endif
		bool status = false;
		status = status || !iod.seekg(static_cast<std::basic_istream<char>::off_type>(pos), ios::beg);
		status = status || !iod.read(&buf[c], static_cast<streamsize>(siz));
		if(status) {
			console::write((boost::format("Can't read diffile at 0x%016X.\n") % p).str(), true);
			return true;
		}
		c += siz;
	}
	return false;
}
bool					device::chgfile::	write(streamptr p, const string &buf) {
	#if !defined NDEBUG && defined DBG_DIFF
		dbglog((boost::format("=>  Write diffile at 0x%016X(%d)\n") % p % buf.size()).str());
	#endif
	auto b = lower_bound(p); // *b >= p
	if(b != begin()) {
		b--;						// *b < p && *b + b->s > p
		if(b->first + b->second.size <= p)
			b++;
	}
	auto e = lower_bound(p + buf.size());	// *e > p + s - 1
	#if !defined NDEBUG && defined DBG_DIFF
		dbglog(
			"	match: " +
			(b == end() ? "end" : (boost::format("0x%016X[0x%016X(%d)]") % b->second.position % b->first % b->second.size).str()) +
			" - " +
			(e == end() ? "end" : (boost::format("0x%016X[0x%016X(%d)]") % e->second.position % e->first % e->second.size).str()) +
			"\n"
		);
	#endif
	auto l = p;
	for(auto i = b; i != e; i++) {
		if(i->first > l) {
			if(addseg(l, buf.substr(l - p, i->first - l)))
				return true;
			l = i->first;
		}
		auto pos = i->second.position;
		if(i->first < p)
			pos += p - i->first;
		auto siz = min<filesize>(buf.size() - (l - p), i->second.size - (i->first < p ? p - i->first : 0));
		#if !defined NDEBUG && defined DBG_DIFF
			dbglog((boost::format("... write diffile at 0x%016X [0x%016X(%d)]\n") % pos % l % siz).str());
		#endif
		bool status = false;
		status = status || !iod.seekp(static_cast<std::basic_istream<char>::off_type>(pos), ios::beg);
		status = status || !iod.write(&buf[l - p], static_cast<streamsize>(siz));
		if(status) {
			console::write((boost::format("Can't write diffile at 0x%016X.\n") % p).str(), true);
			return true;
		}
		l += siz;
	}
	if(l != p + buf.size())
		if(addseg(l, buf.substr(l - p, p + buf.size() - l)))
			return true;
	return false;
}
bool					device::chgfile::	addseg(streamptr p, const string& buf) {
	streamptr s = buf.size();
	device::segment seg;
	seg.size = s;
	bool status = false;
	iod.close();
	iod.open(fatx_context::get()->mmi.diffile, ios::binary | ios::out | ios::app);
	status = status || !iod;
	status = status || !iod.write(reinterpret_cast<char*>(&p), sizeof(p));
	status = status || !iod.write(reinterpret_cast<char*>(&s), sizeof(s));
	seg.position = static_cast<streamptr>(iod.tellp());
	status = status || !iod.write(&buf[0], static_cast<streamsize>(s));
	iod.close();
	status = status || !(iod.open(fatx_context::get()->mmi.diffile, ios::binary | ios::out | ios::in), iod);
	if(status) {
		console::write((boost::format("Can't add segment in diffile at 0x%016X.\n") % p).str(), true);
		return true;
	}
	#if !defined NDEBUG && defined DBG_DIFF
		dbglog((boost::format("... add segment at 0x%016X [0x%016X(%d)]\n") % seg.position % p % s).str());
	#endif
	insert(make_pair(p, seg));
	return false;
}

							device::		device() :
	io(nullptr), iod(nullptr),
	tot_size(0), changes(false), authd("DEV"), chgf() {
}
							device::		~device() {
	if(io)
		io.close();
	if(iod)
		iod.close();
	for(auto &i: usbd)
		i.second.close();
	usbd.clear();
}
string						device::		read(streamptr p, size_t s) {
	if(s == 0)
		return string();
	if(size() && p + s > size()) {
		console::write((boost::format("Blocks out of bounds ([0x%016X ; 0x%016X] > 0x%016X).\n")
			% p
			% (p + s - 1)
			% size()
		).str(), true);
		return string();
	}
	string res(s, '\0');
	bool status = false;
	authd.lock();
	if(usbd.empty()) {
		status = status || !io.seekg(static_cast<std::basic_istream<char>::off_type>(p), ios::beg);
		status = status || !io.read(&res[0], static_cast<streamsize>(s));
		if(status)
			io.clear();
	}
	else {
		auto b = usbd.lower_bound(p);		// *b >= p
		if(b != usbd.begin() && b->first > p)
			b--;						// *b < p && *b + b->s > p
		auto e = usbd.lower_bound(p + s);	// *e > p + s - 1
		streamptr c = 0;
		for(auto i = b; i != e; i++) {
			auto pos = p > i->first ? p - i->first : 0;
			auto siz = min<filesize>(res.size() - c, (next(i, 1) == usbd.end() ? size() : next(i, 1)->first) - i->first - (p > i->first ? pos : 0));
			status = status || !i->second.seekg(static_cast<std::basic_istream<char>::off_type>(pos), ios::beg);
			status = status || !i->second.read(&res[c], static_cast<streamsize>(siz));
			if(status) {
				i->second.clear();
				break;
			}
			c += siz;
		}
	}
	if(!iod)
		status = status || chgf.read(p, s, res);
	if (status) {
		console::write((boost::format("Unreadable block at 0x%016X.\n") % p).str(), true);
		authd.unlock();
		return string();
	} else {
		#if !defined NDEBUG && defined DBG_READ
			devlog(true, p, res);
		#endif
	}
	authd.unlock();
	return res;
}
int							device::		write(streamptr p, const string &s) {
	if(s.empty())
		return 0;
	if(p + s.size() > size()) {
		console::write((boost::format("Blocks out of bounds ([0x%016X;0x%016X] > 0x%016X).\n") % p % (p + s.size() - 1) % size()).str(), true);
		return EOVERFLOW;
	}
	if(!fatx_context::get()->mmi.writeable())
		return 0;
	bool status = false;
	authd.lock();
	#ifndef NO_WRITE
		if(iod)
			status = status || chgf.write(p, s);
		else
			if(usbd.empty()) {
				status = status || !io.seekp(static_cast<std::basic_istream<char>::off_type>(p), ios::beg);
				status = status || !io.write(&s[0], static_cast<streamsize>(s.size()));
				if(status)
					io.clear();
			}
			else {
				auto b = usbd.lower_bound(p);		// *b >= p
				if(b != usbd.begin() && b->first > p)
					b--;						// *b < p && *b + b->s > p
				auto e = usbd.lower_bound(p + s.size());	// *e > p + s - 1
				streamptr c = 0;
				for(auto i = b; i != e; i++) {
					auto pos = p > i->first ? p - i->first : 0;
					auto siz = min<filesize>(s.size() - c, (next(i, 1) == usbd.end() ? size() : next(i, 1)->first) - i->first - (p > i->first ? pos : 0));
					status = status || !i->second.seekg(static_cast<std::basic_istream<char>::off_type>(pos), ios::beg);
					status = status || !i->second.write(&s[c], static_cast<streamsize>(siz));
					if(status) {
						i->second.clear();
						break;
					}
					c += siz;
				}
			}
		changes = true;
	#endif
	if(status) {
		console::write((boost::format("Unwriteable block at 0x%016X.\n") % p).str(), true);
		authd.unlock();
		return EIO;
	}
	else {
		#if !defined NDEBUG && defined DBG_WRITE
			devlog(false, p, s);
		#endif
	}
	authd.unlock();
	return 0;
}
int							device::		setup() {
	bool err = false;
	if(fatx_context::get()->mmi.table != "usb") {
		err = err || !(io = fstream(fatx_context::get()->mmi.input, ios::binary | (
			fatx_context::get()->mmi.writeable() && fatx_context::get()->mmi.diffile.empty() ? (ios::out | ios::in) : ios::in
		)), io);
		err = err || !io.seekg(0);
		err = err || !io.seekg(0, ios::end);
		tot_size = err ? 0 : static_cast<streamptr>(io.tellg());
		if(fatx_context::get()->mmi.writeable() && fatx_context::get()->mmi.diffile.empty()) {
			err = err || !io.seekp(0);
			err = err || !io.seekp(0, ios::end);
		}
	}
	else {
		auto path = fatx_context::get()->mmi.input + sepdir;
		streamptr p = 0;
		fstream f;
		if(f = fstream(path + usb_data + "0000", ios::binary | ios::in), f)
			path = path + usb_data;
		else {
			path = path + usb_dir + sepdir + usb_data;
			err = err || !(f = fstream(path + "0000", ios::binary | ios::in), f);
		}
		err = err || !f.seekg(static_cast<std::basic_istream<char>::off_type>(0x240), ios::beg);
		char buf[8];
		err = err || !f.read(&buf[0], static_cast<streamsize>(8));
		streamptr dts = err ? 0 : byte_order<8>::litend(&buf[0])();
		#if !defined NDEBUG && defined DBG_INIT
			dbglog((boost::format("USB Data of size: %u.\n") % dts).str());
		#endif
		if(f)
			f.close();
		size_t i = 0;
		streamptr s = 0;
		streamptr ts = 0;
		streamptr cts = 0;
		while(!err && (i <= 1 || p < dts)) {
			sprintf(&buf[0], "%04lu", i);
			err = err || !(f = fstream(path + buf, ios::binary | (
				fatx_context::get()->mmi.writeable() && fatx_context::get()->mmi.diffile.empty() ? (ios::out | ios::in) : ios::in
			)), f);
			err = err || !f.seekg(0);
			err = err || !f.seekg(0, ios::end);
			if(fatx_context::get()->mmi.writeable() && fatx_context::get()->mmi.diffile.empty()) {
				err = err || !f.seekp(0);
				err = err || !f.seekp(0, ios::end);
			}
			if(!err) {
				s = static_cast<streamptr>(f.tellg());
				usbd.emplace(p, std::move(f));
				ts += s;
				if(i > 1)
					cts += s;
			}
			#if !defined NDEBUG && defined DBG_INIT
				dbglog((boost::format("USB Data file %s at %u -> %u (%u).\n") % (path + buf) % p % (p + s) % s).str());
			#endif
			p += err ? 0 : s;
			i++;
		}
		err = err || dts != cts;
		tot_size = ts;
	}
	if(!err && !fatx_context::get()->mmi.diffile.empty()) {
		err = err || !(iod = fstream(fatx_context::get()->mmi.diffile, ios::binary | (fatx_context::get()->mmi.writeable() ? (ios::out | ios::in) : ios::in)), iod);
		#if !defined NDEBUG && defined DBG_INIT
			dbglog((boost::format("Diff file requested in %s mode.\n") % (fatx_context::get()->mmi.writeable() ? "write" : "read")).str());
		#endif
		err = err || !iod.seekg(0);
		err = err || !iod.seekg(0, ios::end);
		if(fatx_context::get()->mmi.writeable()) {
			err = err || !iod.seekp(0);
			err = err || !iod.seekp(0, ios::end);
		}
		err = err || chgf.load();
	}
	if(err) {
		if(io)
			io.close();
		if(iod)
			iod.close();
		for(auto &i: usbd)
			i.second.close();
		usbd.clear();
		console::write((boost::format("Error opening %s%s for read%s\n")
			% (fatx_context::get()->mmi.input)
			% (fatx_context::get()->mmi.diffile.empty() ? "" : (":" + fatx_context::get()->mmi.diffile))
			% (fatx_context::get()->mmi.writeable() ? "/write" : "")
		).str(), true);
	}
	#if !defined NDEBUG && defined DBG_INIT
		dbglog("::EODEV\n");
	#endif
	return err ? EIO : 0;
}
#ifndef NDEBUG
string						device::		address(streamptr p) const {
	return (boost::format("0x%016X[%s:0x%08X]")
		% p
		% ((p < fatx_context::get()->par.fat_start) ? "PAR" : (p < fatx_context::get()->par.root_start) ? "FAT" : "CLS")
		% ((p < fatx_context::get()->par.fat_start) ? p : (p < fatx_context::get()->par.root_start) ? ((p - fatx_context::get()->par.fat_start) / fatx_context::get()->par.chain_size) : clsarithm::ptr2cls(p))
	).str();
}
void						device::		devlog(bool r, streamptr p, const string& s) const {
	dbglog((boost::format("->  %s at %s (%d) :%s")
		% (r ? "read" : "write")
		% address(p)
		% s.size()
		% ((s.size() <= DBGCR) ? "" : "\n")
	).str());
	string sl;
	for(
		size_t i = 0;
		i <
			#ifdef DBGLIMIT
				min<string::size_type>(string::size_type(DBGLIMIT), s.size());
			#else
				s.size();
			#endif
		i++
	) {
		sl += (boost::format(" %02X") % static_cast<unsigned int>(static_cast<unsigned char>(s[i]))).str();
		if(((i + 1) % DBGCR) == 0) {
			dbglog(sl + "\n");
			sl.clear();
		}
	}
	dbglog(sl + "\n");
}
string						device::		print(streamptr p, size_t s, size_t g) {
	string res;
	string buf;
	size_t i = 1;
	for(const char c: read(p, s)) {
		buf += c;
		if((i % g) == 0) {
			for(size_t j = 0; j < g; j++)
				res += (boost::format("%02X ") % static_cast<unsigned int>(static_cast<unsigned char>(buf[j]))).str();
			for(size_t j = 0; j < g; j++)
				res += (buf[j] >= ' ' && buf[j] <= '~') ? buf[j] : '.';
			res += '\n';
			buf.erase();
		}
		i++;
	}
	return res;
}
#endif

							fatxpar::		fatxpar() :
	par_id(0),				par_label(),			par_start(0),			par_size(0),			clus_size(0),			clus_pow(0),
	clus_num(0),			clus_fat(0),			chain_size(0),			chain_pow(0),			fat_start(0),			fat_size(0),
	root_start(0),			root_clus(0) { }
int							fatxpar::		setup() {
	uint64_t ts = fatx_context::get()->dev.size();
	bool found = false;
	if(fatx_context::get()->mmi.verbose)
		console::write((boost::format("Support size: %d.\n") % ts).str());
	map<const string, map<const string, const pair<streamptr, streamptr>>> sch = {
		{ "file", {
			{ "x2",		{	0x0,			0x0			}}
		}},
		{ "mu", {
			{ "sc",		{	0x0,			0x7FF000	}},
			{ "x2",		{	0x7FF000,		0x0			}}
		}},
		{ "usb", {
			{ "sc",		{	0x8000400,		0x12000400	}},
			{ "se1",	{	0x8115200,		0x8000000	}},
			{ "se2",	{	0x12000400,		0xDFFFC00	}},
			{ "x2",		{	0x20000000,		0x0			}}
		}},
		{ "hd", {
			{ "sc",		{	0x80000,		0x80000000	}},
			{ "gc",		{	0x80080000,		0xA0E30000	}},
			{ "se1",	{	0x10C080000,	0xCE30000	}},
			{ "se2",	{	0x118EB0000,	0x8000000	}},
			{ "x1",		{	0x120eb0000,	0x10000000	}},
			{ "x2",		{	0x130eb0000,	0x0			}}
		}}
	};
	map<const string, const string> names = {
		{ "file",	"plain file" },
		{ "mu",		"memory unit" },
		{ "usb",	"USB drive" },
		{ "hd",		"Xbox 360 retail hard disk" },
		{ "kit",	"devkit hard disk" },
		{ "sc",		"system cache" },
		{ "gc",		"game cache" },
		{ "se1",	"sysext" },
		{ "se2",	"sysext 2" },
		{ "xdv",	"Xbox 360 dashboard"},
		{ "x1",		"original Xbox compatibility" },
		{ "x2",		"data" }
	};
	map<const string, set<string>> cap;
	for(const auto &i: sch) {
		if(ts <= ranges::find_if(i.second, [] (const auto j) noexcept { return j.second.second == 0; })->second.first)
			continue;
		cap.emplace(i.first, set<string>{});
		for(const auto &j: i.second) {
			if(fatx_context::get()->dev.read(j.second.first).find(fsid, 0) == 0) {
				cap[i.first].insert(j.first);
				if(fatx_context::get()->mmi.verbose)
					console::write((boost::format("Found FATX filesystem in %s partition in %s table.\n")
						% names[j.first]
						% names[i.first]
					).str());
			}
		}
	}
	if(cap.find("mu") != cap.end() && cap.find("mu")->second.size() == sch["mu"].size() && cap.find("file") != cap.end())
		cap.erase("file");
	#ifndef NDEBUG
		dbglog("CAPACITY:\n");
		for(const auto &i: cap) {
			dbglog("  " + i.first + "\n");
			for(const auto &j: i.second)
				dbglog("    " + j + "\n");
		}
	#endif
	if(
		fatx_context::get()->mmi.offset != 0 &&
		(fatx_context::get()->mmi.prog == frontend::mkfs || fatx_context::get()->dev.read(fatx_context::get()->mmi.offset).find(fsid, 0) == 0)
	) {
		par_start = fatx_context::get()->mmi.offset;
		par_size = ts - par_start;
		if(fatx_context::get()->mmi.verbose && fatx_context::get()->dev.read(fatx_context::get()->mmi.offset).find(fsid, 0) == 0)
			console::write((boost::format("Found FATX partition at 0x%016X.\n") % par_start).str());
		found = fatx_context::get()->dev.read(fatx_context::get()->mmi.offset).find(fsid, 0) == 0;
	}
	else if(
		cap.find(fatx_context::get()->mmi.table) != cap.end() && (
			fatx_context::get()->mmi.prog == frontend::mkfs ||
			cap[fatx_context::get()->mmi.table].find(fatx_context::get()->mmi.partition) != cap[fatx_context::get()->mmi.table].end()
		)
	) {
		par_start = sch[fatx_context::get()->mmi.table][fatx_context::get()->mmi.partition].first;
		par_size = sch[fatx_context::get()->mmi.table][fatx_context::get()->mmi.partition].second != 0 ? sch[fatx_context::get()->mmi.table][fatx_context::get()->mmi.partition].second : ts - par_start;
		found = cap[fatx_context::get()->mmi.table].find(fatx_context::get()->mmi.partition) != cap[fatx_context::get()->mmi.table].end();
	}
	else if(fatx_context::get()->mmi.table == "kit") {
		devheader dh(fatx_context::get()->dev.read(0).data());
		if(dh.id != 0x00020000 && fatx_context::get()->mmi.prog == frontend::mkfs)
			dh = devheader(ts);
		if(fatx_context::get()->dev.read(dh.p2_start * blksize).find(fsid, 0) == 0) {
			if(fatx_context::get()->mmi.verbose)
				console::write((boost::format("Found FATX filesystem in %s partition in %s table.\n")
					% names["x2"]
					% names["kit"]
				).str());
			if(fatx_context::get()->mmi.partition == "x2") {
				par_start	= dh.p2_start * blksize;
				par_size	= dh.p2_size * blksize;
				found = true;
			}
		}
		if(fatx_context::get()->dev.read(dh.p1_start * blksize).find(fsid, 0) == 0) {
			if(fatx_context::get()->mmi.verbose)
				console::write((boost::format("Found FATX filesystem in %s partition in %s table.\n")
					% names["xdv"]
					% names["kit"]
				).str());
			if(fatx_context::get()->mmi.partition == "xdv") {
				par_start	= dh.p1_start * blksize;
				par_size	= dh.p1_size * blksize;
				found = true;
			}
		}
	}
	else {
		if(fatx_context::get()->mmi.prog == frontend::mkfs) {
			console::write((boost::format("No space for partition in %s table.\n") % names[fatx_context::get()->mmi.table]).str(), true);
			return ENOSPC;
		}
		else {
			console::write("No FATX partition found.\n", true);
			return ENODATA;
		}
	}
	if(fatx_context::get()->mmi.size != 0) {
		if(fatx_context::get()->mmi.offset == 0 && fatx_context::get()->mmi.table != "file")
			console::write((boost::format("Can't force size for a partition in %s table. Ignoring.\n") % names[fatx_context::get()->mmi.partition]).str(), true);
		else if(fatx_context::get()->mmi.size >= par_size)
			console::write((boost::format("Can't force a greater size than possible (limit is %d). Ignoring.\n") % par_size).str(), true);
		else
			par_size = fatx_context::get()->mmi.size;
	}
	if(found) {
		if(fatx_context::get()->mmi.verbose)
			console::write((boost::format("Using %s partition in %s table.\n") % names[fatx_context::get()->mmi.partition] % names[fatx_context::get()->mmi.table]).str());
		bootsect	bs(fatx_context::get()->dev.read(par_start).data());
		par_id		= bs.id;
		root_clus	= bs.root;
		clus_size	= static_cast<uint32_t>(blksize * (fatx_context::get()->mmi.clus_size ? fatx_context::get()->mmi.clus_size : (bs.spc == 0 || bs.spc > 0xFFFF) ? 1 : bs.spc));
	}
	else {
		par_id		= 0;
		root_clus	= 1;
		clus_size	= static_cast<uint32_t>(blksize * (fatx_context::get()->mmi.clus_size ? fatx_context::get()->mmi.clus_size : (
			par_size > 0x200000000ULL ?	512	:
			par_size > 0x100000000ULL ?	256	:
			par_size > 0x080000000ULL ?	128	:
			par_size > 0x040000000ULL ?	64	:
			par_size > 0x020000000ULL ?	32	:
			par_size > 0x010000000ULL ?	16	:
			par_size > 0x008000000ULL ?	8	:
			par_size > 0x001000000ULL ?	4	:
			par_size > 0x000800000ULL ?	8	:
			par_size > 0x000400000ULL ?	4	:
			par_size > 0x000200000ULL ?	2	:
										1
		)));
	}
	string s(bitset<64>(clus_size).to_string());
	if(s.find('1') != s.rfind('1')) {
		console::write("Size of clusters is not a power of 2.\n", true);
		return EINVAL;
	}
	clus_pow	= static_cast<unsigned short>(63 - s.find('1'));
	clus_num	= static_cast<uint32_t>(par_size >> clus_pow);
	chain_size	= clus_num < 0xFFF0 ? 2 : 4;
	chain_pow	= chain_size == 2 ? 1 : 2;
	fat_start	= par_start + 0x1000;
	fat_size	= clus_num * chain_size;
	fat_size	+= (0x1000 - (fat_size % 0x1000));
	root_start	= fat_start + fat_size;
	clus_fat	= static_cast<uint32_t>(((par_size - (root_start - par_start)) >> clus_pow) - 1);
	if(root_clus < 1 || root_clus > clus_fat) {
		root_clus = 1;
		if(fatx_context::get()->mmi.prog != frontend::mkfs) {
			console::write("Bad root cluster number.", fatx_context::get()->mmi.dialog);
			if(fatx_context::get()->mmi.prog == frontend::fsck) {
				console::write(" Correct it ?", fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.getanswer(true))
					write();
			}
			else
				console::write("\n", fatx_context::get()->mmi.dialog);
		}
	}
	#ifndef NDEBUG
		dbglog((boost::format("PAR size  : %d\n") % par_size).str());
		dbglog((boost::format("CLS size  : %d (%d)\n") % clus_size % clus_pow).str());
		dbglog((boost::format("CLS num   : %d\n") % clus_num).str());
		dbglog((boost::format("FAT start : 0x%016X\n") % fat_start).str());
		dbglog((boost::format("FAT cls   : %d\n") % clus_fat).str());
		dbglog((boost::format("ROOT start: 0x%016X (0x%08X)\n") % root_start % root_clus).str());
	#endif
	#if !defined NDEBUG && defined DBG_INIT
		dbglog("::EOPAR\n");
	#endif
	return 0;
}
int							fatxpar::		write() {
	if(fatx_context::get()->mmi.table == "kit") {
		int res;
		string buf(blksize, '\0');
		devheader(fatx_context::get()->dev.size()).write(&buf[0]);
		if((res = fatx_context::get()->dev.write(0, buf)))
			return res;
	}
	string buf(blksize, '\0');
	bootsect(par_id, static_cast<uint32_t>(clus_size / blksize), static_cast<uint32_t>(root_clus)).write(&buf[0]);
	return fatx_context::get()->dev.write(par_start, buf);
}
size_t						fatxpar::		label(unsigned char buf[slab]) const {
	size_t res = 2;
	memset(buf, 0, slab);
	buf[0] = 0xFE;
	buf[1] = 0xFF;
	for(size_t i = 0; i < par_label.size(); i++) {
		if(res == slab)
			break;
		res += 2;
		buf[res - 1] = static_cast<unsigned char>(par_label[i]);
	}
	return res;
}
void						fatxpar::		label(const unsigned char buf[slab], const size_t size) {
	for(size_t i = 3; i < slab; i += 2) {
		if(i >= size)
			break;
		par_label += static_cast<char>(buf[i]);
	}
}

clusptr						clsarithm::		siz2cls(filesize s) {
	return (s >> fatx_context::get()->par.clus_pow)
		   + (((s - ((s >> fatx_context::get()->par.clus_pow) << fatx_context::get()->par.clus_pow))
			   != 0)
				  ? 1
				  : 0);
}
clusptr						clsarithm::		inccls(clusptr p) {
	return (p <= fatx_context::get()->par.clus_fat) ? p + 1 : 2;
}
streamptr					clsarithm::		cls2ptr(clusptr p) {
	if (p < fatx_context::get()->par.root_clus || p > fatx_context::get()->par.clus_fat) {
		console::write((boost::format("Cluster pointer in data out of bounds (0x%08X).\n") % p).str(), true);
		return 0;
	}
	return fatx_context::get()->par.root_start + (p - 1) * fatx_context::get()->par.clus_size;
}
clusptr						clsarithm::		ptr2cls(streamptr p) {
	return ((p - fatx_context::get()->par.root_start) >> fatx_context::get()->par.clus_pow) + 1;
}
streamptr					clsarithm::		cls2fat(clusptr p) {
	if (p < fatx_context::get()->par.root_clus || p > fatx_context::get()->par.clus_fat) {
		console::write((boost::format("Cluster pointer in fat out of bounds (0x%08X).\n") % p).str(), true);
		return 0;
	}
	return fatx_context::get()->par.fat_start + p * fatx_context::get()->par.chain_size;
}
string						clsarithm::		clsprint(clusptr p, clusptr r) {
	return (p == r + 1)
			   ? "next"
			   : ((p == FLK) ? "free" : ((p == EOC) ? "end" : (boost::format("0x%08X") % p).str()));
}

							dskmap::		dskmap(const fatxpar& par) :
	memnext(
		std::bind(&dskmap::real_read, this, std::placeholders::_1, std::placeholders::_2),
		std::bind(&dskmap::real_write, this, std::placeholders::_1, std::placeholders::_2),
		#define CACHESIZE (par.clus_fat * par.chain_size / max_cache_div > par.clus_size ? par.clus_fat * par.chain_size / max_cache_div : par.clus_size)
		CACHESIZE,
		(CACHESIZE / nb_cache_div > par.clus_size ? CACHESIZE / nb_cache_div : par.clus_size)
		#undef CACHESIZE
	), authm("FAT") { }
							dskmap::		~dskmap() {
	memnext.clear();
}
void						dskmap::		forfat(const lbdfat_t &lbd) {
	clusptr c = fatx_context::get()->par.root_clus;
	for(
		streamptr p = fatx_context::get()->par.fat_start;
		p < fatx_context::get()->par.fat_start + fatx_context::get()->par.fat_size;
		p += fatx_context::get()->par.clus_size
	) {
		string buf = fatx_context::get()->dev.read(p, fatx_context::get()->par.clus_size);
		for(
			uint16_t i = static_cast<uint16_t>(p == fatx_context::get()->par.fat_start ? fatx_context::get()->par.root_clus : 0);
			i < (fatx_context::get()->par.clus_size >> fatx_context::get()->par.chain_pow) && c < fatx_context::get()->par.clus_fat;
			i++, c++
		)
			lbd(c, (
				fatx_context::get()->par.chain_size == 4) ?
				byte_order<4>::litend(&buf[static_cast<string::size_type>(i * fatx_context::get()->par.chain_size)])() :
				byte_order<2>::litend(&buf[static_cast<string::size_type>(i * fatx_context::get()->par.chain_size)])()
			);
	}
}
dskmap::memnext_t::lkval_t  dskmap::		real_read(clusptr p, size_t s) {
	memnext_t::lkval_t res;
	s = min<size_t>(s, fatx_context::get()->par.clus_fat - p);
	string buf = fatx_context::get()->dev.read(clsarithm::cls2fat(p), fatx_context::get()->par.chain_size * s);
	for(size_t i = 0; i < buf.size(); i += fatx_context::get()->par.chain_size) {
		clusptr a = (fatx_context::get()->par.chain_size == 4) ? byte_order<4>::litend(&buf[i])() : byte_order<2>::litend(&buf[i])();
		if(fatx_context::get()->par.chain_size == 2 && a == (EOC & 0xFFFF))
			a = EOC;
		if(a != FLK && a != EOC && a > fatx_context::get()->par.clus_fat && bad.find(p + i) == bad.end()) {
			console::write((boost::format("Cluster value in FAT out of bounds (0x%08X) for cluster 0x%08X.") % a % (p + i)).str(), fatx_context::get()->mmi.dialog);
			bad.insert(p + i);
			if(fatx_context::get()->mmi.prog == frontend::fsck) {
				console::write(" Free it ?", fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.getanswer(true)) {
					write((p + i), FLK);
					a = FLK;
				}
			}
			else
				console::write("\n", fatx_context::get()->mmi.dialog);
		}
		res.push_back(memnext_t::pair_type(a, p + (i >> fatx_context::get()->par.chain_pow)));
	}
	return res;
}
int							dskmap::		real_write(clusptr p, clusptr v) {
	string buf(fatx_context::get()->par.chain_size, '\0');
	if(fatx_context::get()->par.chain_size == 4)
		buf = byte_order<4>::litend(static_cast<byte_order<4>::value_type>(v));
	else
		buf = byte_order<2>::litend(static_cast<byte_order<2>::value_type>(v));
	return fatx_context::get()->dev.write(clsarithm::cls2fat(p), buf);
}
vareas						dskmap::		getareas(clusptr orig, const lbdarea_t &lbd) {
	if(lbd == nullptr)
			authm.lock_shared();
	set<clusptr> sc;
	vareas res;
	if(orig == EOC || orig == FLK)
		return res;
	streamptr	area_off	= 0;
	streamptr	area_ptr	= 0;
	streamptr	area_siz	= 0;
	clusptr		area_start	= 0;
	clusptr cur_cls = orig, prv_cls = 0;
	while(true) {
		if(cur_cls != EOC && cur_cls != FLK) {
			if(sc.find(cur_cls) == sc.end())
				sc.insert(cur_cls);
			else {
				console::write((boost::format("Circular reference in FAT chain starting at 0x%08X.") % orig).str(), fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.prog == frontend::fsck) {
					console::write(" Cut it ?", fatx_context::get()->mmi.dialog);
					if(fatx_context::get()->mmi.getanswer(true))
						write(prv_cls, EOC);
				}
				else
					console::write(" Ignoring.\n", fatx_context::get()->mmi.dialog);
				break;
			}
			if(area_ptr == 0) {
				area_ptr = clsarithm::cls2ptr(cur_cls);
				area_start = cur_cls;
			}
			area_siz += fatx_context::get()->par.clus_size;
		}
		if((prv_cls != 0 && prv_cls != cur_cls - 1) || cur_cls == EOC || cur_cls == FLK) {
			if(cur_cls != EOC && cur_cls != FLK)
				area_siz -= fatx_context::get()->par.clus_size;
			res.push_back(area(area_off, area_ptr, area_siz, area_start, prv_cls));
			area_off += area_siz;
			if(cur_cls != EOC && cur_cls != FLK)
				area_ptr = clsarithm::cls2ptr(cur_cls);
			area_siz = fatx_context::get()->par.clus_size;
			area_start = cur_cls;
		}
		if(cur_cls == EOC || cur_cls == FLK)
			break;
		cur_cls = read(prv_cls = cur_cls);
		if(lbd)
			lbd(prv_cls, cur_cls);
	}
	#if !defined NDEBUG && defined DBG_AREAS
		dbglog((lbd ? "Write" : "Get") + res.print());
	#endif
	if(lbd == nullptr)
			authm.unlock_shared();
	return res;
}
clusptr						dskmap::		clsavail() {
	clusptr res = 0;
	if(freegaps.empty())
		gapcheck();
	for(const auto& i: freegaps.right)
		res += i.first;
	return res;
}
void						dskmap::		erase() {
	authm.lock();
	freegaps.clear();
	fatx_context::get()->dev.write(clsarithm::cls2fat(fatx_context::get()->par.root_clus), string(
		(fatx_context::get()->par.clus_fat - fatx_context::get()->par.root_clus) * fatx_context::get()->par.chain_size,
		'\0'
	));
	memnext.clear();
	gapcheck();
	authm.unlock();
}
void						dskmap::		gapcheck() {
	authm.lock();
	#ifndef NDEBUG
		dbglog((boost::format("Calculating free gaps out of %d FAT entries...\n") % fatx_context::get()->par.clus_fat).str());
	#endif
	freegaps.clear();
	mapptr_t b = 0;
	mapsiz_t s = 0;
	forfat([this, &b, &s](clusptr o, clusptr v) -> void {
		if(v == FLK) {
			if(b != 0)
				s++;
			else {
				b = o;
				s = 1;
			}
		}
		else if(b != 0) {
			freegaps.insert(gap_t::value_type(b, s));
			b = 0;
		}
	});
	if(b != 0)
		freegaps.insert(gap_t::value_type(b, s));
	#if !defined NDEBUG && defined DBG_GAPS
		dbglog("Gaps:\n");
		printgaps();
	#endif
	authm.unlock();
}
clusptr						dskmap::		read(clusptr p) {
	if(p == FLK || p == EOC) {
		console::write((boost::format("Can't read FAT at special cluster value (0x%08X).\n") % p).str(), true);
		return 0;
	}
	if(p > fatx_context::get()->par.clus_fat) {
		console::write((boost::format("Cluster pointer to FAT out of bounds (0x%08X).\n") % p).str(), true);
		return 0;
	}
	return memnext(p);
}
int							dskmap::		write(clusptr p, clusptr v) {
	if (p == FLK || p == EOC) {
		console::write((boost::format("Can't write FAT at special cluster value (0x%08X).\n") % p).str(), true);
		return EOVERFLOW;
	}
	if(p > fatx_context::get()->par.clus_fat) {
		console::write((boost::format("Cluster pointer to FAT out of bounds (0x%08X).\n") % p).str(), true);
		return EOVERFLOW;
	}
	if(v != FLK && v != EOC && v > fatx_context::get()->par.clus_fat) {
		console::write((boost::format("Cluster value to FAT out of bounds (0x%08X) for cluster 0x%08X.\n") % v % p).str(), true);
		return EOVERFLOW;
	}
	return memnext(p, v);
}
vareas						dskmap::		alloc(clusptr s, clusptr o) {
	vareas res;
	if(s == 0)
		return vareas();
	authm.lock();
	if(freegaps.empty()) {
		console::write("No space left on device, disk full.\n", true);
		authm.unlock();
		return vareas();
	}
	clusptr			gap_clus = 0;
	clusptr			gap_size = 0;
	if(o != 0) {
		// we first try to allocate in continuity with o
		gap_t::left_map::iterator			prev = freegaps.left.find(o);
		if(prev != freegaps.left.end() && prev->second >= s) {
			gap_clus	= prev->first;
			gap_size	= prev->second;
		}
	}
	if(gap_clus == 0) {
		// we try to allocate at end of used cluster
		gap_t::left_map::reverse_iterator	end = freegaps.left.rbegin();
		if(end != freegaps.left.rend() && end->second >= s) {
			gap_clus	= end->first;
			gap_size	= end->second;
		}
	}
	if(gap_clus == 0) {
		// we find smallest gap that fits
		gap_t::right_map::iterator			fit = freegaps.right.lower_bound(s);
		if(fit != freegaps.right.end() && fit->first == s) {
			gap_clus	= fit->second;
			gap_size	= fit->first;
		}
	}
	if(gap_clus != 0) {
		// contiguous case
		for(clusptr i = gap_clus; i < gap_clus + s; i++)
			write(i, (i == gap_clus + s - 1) ? EOC : i + 1);
		freegaps.left.erase(gap_clus);
		if(gap_size != s)
			freegaps.insert(gap_t::value_type(gap_clus + s, gap_size - s));
		res.push_back(area(
			0,
			clsarithm::cls2ptr(gap_clus),
			s * fatx_context::get()->par.clus_size,
			gap_clus,
			gap_clus + s - 1
		));
	}
	else {
		// first, evaluate the number of free clusters
		gap_size	= 0;
		for(const auto& i: freegaps.right)
			gap_size += i.first;
		if(gap_size >= s) {
			// we take gaps in decreasing size order
			gap_t::right_map::reverse_iterator gap;
			clusptr	tot_size = s;
			clusptr old_clus = 0;
			do {
				gap = freegaps.right.rbegin();
				assert(gap != freegaps.right.rend());
				gap_clus = gap->second;
				gap_size = gap->first;
				if(old_clus != 0)
					write(old_clus, gap_clus);
				for(clusptr i = gap_clus; i < gap_clus + min<clusptr>(gap_size, tot_size); i++)
					write(i, (i == gap_clus + min<clusptr>(gap_size, tot_size) - 1) ? EOC : i + 1);
				res.push_back(area(
					res.empty() ? 0 : res.back().offset + res.back().size,
					clsarithm::cls2ptr(gap_clus),
					min<clusptr>(gap_size, tot_size) * fatx_context::get()->par.clus_size,
					gap_clus,
					gap_clus + min<clusptr>(gap_size, tot_size) - 1
				));
				freegaps.right.erase(gap_clus);
				if(tot_size < gap_size) {
					freegaps.insert(gap_t::value_type(gap_clus + gap_size, tot_size - gap_size));
					tot_size = 0;
				}
				else
					tot_size -= gap_size;
				old_clus = gap_clus;
			} while(tot_size != 0);
		}
		else {
			console::write((boost::format("Not enough disk space for %d cluster allocation.\n") % s).str(), true);
			authm.unlock();
			return vareas();
		}
	}
	#ifndef NDEBUG
		dbglog(
			gap_clus ? (
				boost::format("*** FAT alloc: %d cluster%s starting at 0x%08X.\n") % s % (s > 1 ? "s" : "") % gap_clus
			).str() : (
				boost::format("*** FAT alloc: failed to allocate %d clusters.\n") % s
			).str()
		);
	#endif
	authm.unlock();
	return res;
}
void						dskmap::		freefat(clusptr o) {
	if(o == FLK || o == EOC)
		return;
	authm.lock();
	vareas va = getareas(o, [this](clusptr c, clusptr) -> void { write(c, FLK); });
	if(fatx_context::get()->mmi.prog == frontend::fsck) {
		authm.unlock();
		return;
	}
	for(area i: va) {
		#ifndef NDEBUG
			dbglog((boost::format("*** FAT free:  %d cluster%s starting at 0x%08X.\n") % (i.stop - i.start + 1) % ((i.stop - i.start) > 0 ? "s" : "") % i.start).str());
		#endif
		gap_t::left_map::iterator next = freegaps.left.upper_bound(i.start);
		gap_t::left_map::iterator prev = next;
		--prev;
		if(next != freegaps.left.end() && prev->first + prev->second == i.start && i.stop + 1 == next->first) {
			// new gap is adjascent with previous and next gap
			clusptr prev_clus = prev->first;
			clusptr prev_size = prev->second;
			clusptr next_size = next->second;
			freegaps.left.erase(prev);
			freegaps.left.erase(next);
			freegaps.insert(gap_t::value_type(prev_clus, prev_size + i.stop - i.start + 1 + next_size));
		}
		else if(next != freegaps.left.end() && prev->first + prev->second == i.start) {
			// new gap is adjascent with a previous gap
			clusptr prev_clus = prev->first;
			clusptr prev_size = prev->second;
			freegaps.left.erase(prev);
			freegaps.insert(gap_t::value_type(prev_clus, prev_size + i.stop - i.start + 1));
		}
		else if(next != freegaps.left.end() && i.stop + 1 == next->first) {
			// new gap is adjascent with a next gap
			clusptr next_size = next->second;
			freegaps.left.erase(next);
			freegaps.insert(gap_t::value_type(i.start, i.stop - i.start + 1 + next_size));
		}
		else {
			// new gap is not adjascent with another gap
			freegaps.insert(gap_t::value_type(i.start, i.stop - i.start + 1));
		}
	}
	authm.unlock();
}
int							dskmap::		resize(ptr_vareas o, clusptr s) {
	if(!o)
		return EFAULT;
	if(o->empty()) {
		if(s == 0)
			return 0;
		else
			return EFAULT;
	}
	if(s == 0) {
		freefat(o->first());
		return 0;
	}
	bool res = true;
	if(o->nbcls() < s) {
		// we need to extend the chain
		vareas extend = alloc(s - o->nbcls(), o->last() + 1);
		res = !extend.empty();
		if(res) {
			authm.lock();
			res = write(o->last(), extend.first());
			o->add(extend);
			authm.unlock();
		}
	}
	else if(o->nbcls() > s) {
		// we need to reduce the chain
		authm.lock();
		res = write(o->at(s), EOC);
		authm.unlock();
		if(res)
			freefat(o->at(s + 1));
		authm.lock();
		o->erase(o->in(s) + 1, o->end());
		o->back().size -= o->nbcls() - s;
		o->back().stop = o->back().start + o->back().size - 1;
		authm.unlock();
	}
	return res;
}
void						dskmap::		change(clusptr, entry*, clusptr, status_t) {
	console::write(string("Invalid call to :") + __FUNCTION__ + "\n", true);
	exit(2);
}
dskmap::status_t			dskmap::		status(clusptr) const {
	console::write(string("Invalid call to :") + __FUNCTION__ + "\n", true);
	exit(2);
}
entry*						dskmap::		getentry(clusptr) const {
	console::write(string("Invalid call to :") + __FUNCTION__ + "\n", true);
	exit(2);
}
void						dskmap::		fatlost() {
	console::write(string("Invalid call to :") + __FUNCTION__ + "\n", true);
	exit(2);
}
void						dskmap::		fatcheck() {
	console::write(string("Invalid call to :") + __FUNCTION__ + "\n", true);
	exit(2);
}
string						dskmap::		printchain(clusptr orig) {
	string res;
	while(orig != FLK && orig != EOC) {
		res += (boost::format("->0x%08X") % orig).str();
		orig = dskmap::read(orig);
	}
	return res;
}
#ifndef NDEBUG
void						dskmap::		printgaps() const {
	for(const auto& i: freegaps.left)
		dbglog((boost::format("%08X: %d cluster%s free\n") % i.first % i.second % (i.second > 1 ? "s": "")).str());
}
void						dskmap::		printfat() {
	console::write(string("Invalid call to :") + __FUNCTION__ + "\n", true);
	exit(2);
}
#endif

							memmap::		memmap(const fatxpar& par) : dskmap(par) {
}
clusptr						memmap::		read(clusptr p) {
	return (memchain.find(p) != memchain.end()) ? memchain.find(p)->second.next : (fatx_context::get()->mmi.dellost ? dskmap::read(p) : FLK);
}
void						memmap::		change(clusptr p, entry* e, clusptr n, status_t s) {
	if(memchain.find(p) != memchain.end()) {
		if(n != FLK)
			memchain.find(p)->second.next	= n;
		memchain.find(p)->second.ent		= e;
		memchain.find(p)->second.status		= s;
	}
	else
		memchain.insert(make_pair(p, link(e, n, s)));
}
memmap::status_t			memmap::		status(clusptr p) const {
	return (memchain.find(p) == memchain.end()) ? disk : memchain.find(p)->second.status;
}
entry*						memmap::		getentry(clusptr p) const {
	return (memchain.find(p) == memchain.end()) ? nullptr : memchain.find(p)->second.ent;
}
void						memmap::		fatlost() {
	lost.clear();
	set<clusptr> l;
	forfat([this, &l](clusptr o, clusptr v) -> void {
		if(v != FLK && status(o) == disk && l.find(o) == l.end()) {
			auto f = find_if(lost.begin(), lost.end(), [v] (const vareas& i) -> bool { return i.first() == v; });
			if(f != lost.end()) {
				f->add(o);
				l.insert(o);
			}
			else {
				vareas va = getareas(o, [&l] (const clusptr i, const clusptr) -> void { l.insert(i); });
				lost.push_back(va);
			}
		}
	});
}
void						memmap::		fatcheck() {
	if(fatx_context::get()->mmi.prog == frontend::fsck) {
		// we propose to correct each erroneous entry in fat
		for(const auto& p: memchain) {
			if(status(p.first) == modified) {
				console::write((boost::format("Cluster number in FAT 0x%08X shall be %s instead of %s. Correct it ?")
					% p.first
					% clsarithm::clsprint(p.second.next, p.first)
					% clsarithm::clsprint(dskmap::read(p.first), p.first)
				).str(), fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.getanswer(true)) {
					write(p.first, p.second.next);
					memchain.erase(p.first);
				}
			}
		}
	}
	if(fatx_context::get()->mmi.prog == frontend::fsck || fatx_context::get()->mmi.prog == frontend::unrm) {
		// we propose to rescue or to clear each chain with no reference found in fat
		for(const vareas& va: lost) {
			if(fatx_context::get()->mmi.prog == frontend::fsck) {
				console::write((boost::format("Found unknown chain at 0x%08X (%d). Free it ?")
					% va.first()
					% (va.nbcls() * fatx_context::get()->par.clus_size)
				).str(), fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.getanswer(true))
					freefat(va.first());
			}
			else {
				console::write((boost::format("Found unknown chain at 0x%08X (%d). Recover in %s ?")
					% va.first()
					% (va.nbcls() * fatx_context::get()->par.clus_size)
					% fatx_context::get()->mmi.lostfound
				).str(), fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.getanswer(false)) {
					if(fatx_context::get()->mmi.local) {
						ptr_entry f(new entry((boost::format("%s%03d") % fatx_context::get()->mmi.foundfile % fatx_context::get()->mmi.filecount++).str()));
						f->cluster(va.first());
						f->size(va.nbcls() * fatx_context::get()->par.clus_size);
						f->recover();
					}
					else {
						entry* lf = fatx_context::get()->root->find(fatx_context::get()->mmi.lostfound.data());
						if(lf == nullptr) {
							// we first have to create lost+found directory
							if(fatx_context::get()->root->addtodir((lf = new entry(fatx_context::get()->mmi.lostfound, 0, true)))) {
								console::write((boost::format("Unable to create directory %s.\n") % fatx_context::get()->mmi.lostfound).str(), fatx_context::get()->mmi.dialog);
								return;
							}
						}
						else {
							// we find the latest file number
							for(entry& e: lf->childs) {
								unsigned int n;
								if(sscanf(e.name(), (string(def_fpre) + "%3d").data(), &n) == 1)
									fatx_context::get()->mmi.filecount = max<unsigned int>(fatx_context::get()->mmi.filecount, n + 1);
							}
						}
						// we create the file and put it in lost+found
						entry* f = new entry((boost::format("%s%03d") % fatx_context::get()->mmi.foundfile % fatx_context::get()->mmi.filecount++).str(), 0);
						f->size(va.nbcls() * fatx_context::get()->par.clus_size);
						f->cluster(va.first());
						if(lf->addtodir(f))
							console::write((boost::format("Unable to create file %s.\n") % f->name()).str(), fatx_context::get()->mmi.dialog);
					}
				}
			}
		}
	}
}
#ifndef NDEBUG
void						memmap::		printfat() {
	bool empty = false;
	status_t s = marked;
	entry* ent = fatx_context::get()->root;
	clusptr l = 1;
	for(clusptr i = fatx_context::get()->par.root_clus; i < fatx_context::get()->par.clus_fat; i++) {
		if(
			empty != ((read(i) == FLK) && (dskmap::read(i) == FLK)) ||
			(!empty && (
				(s != status(i)) ||
				(status(i) != disk && ent != getentry(i))
			)) ||
			i == fatx_context::get()->par.clus_fat
		) {
			dbglog((boost::format("%s %s%s\n")
				% (
					!empty ? (
						(s == disk) ? "LOST" :
						(s == marked) ? ((getentry(ent->cluster()) == ent) ? ((ent->cluster() == l) ? "OK  " : "OK->") : "*ERR") :
						(s == deleted) ? ((getentry(ent->cluster()) == ent) ? ((ent->cluster() == l) ? "DEL " : "DEL>") : "*BAD") :
						"?MOD"
					) : "...."
				)
				% (
					(l == (i - 1)) ?
					(boost::format("0x%08X:			") % l).str() :
					(boost::format("0x%08X-0x%08X: ") % l % (i - 1)).str()
				)
				% (
					(!empty && s != disk) ?
					(boost::format("(0x%08X: %s)") % ent->cluster() % ent->path()).str() : ""
				)
			).str());
			empty = (read(i) == FLK) && (dskmap::read(i) == FLK);
			s = status(i);
			ent = (status(i) == disk) ? nullptr : getentry(i);
			l = i;
		}
	}
}
#endif

							// root entry constructor
							entry::			entry() :
	mux_B(string() + mutex_buff + "/"),
	mux_D(string() + mutex_data + "/"),
	mux_E(string() + mutex_entr + "/"),
	entbuf(),
	v_cptacc(0),
	v_writeopened(fatx_context::get()->mmi.prog != frontend::fuse),
	v_status(valid),
	v_namesize(0),
	v_flags(),
	v_cluster(fatx_context::get()->par.root_clus),
	v_size(0),
	v_creation(),
	v_access(),
	v_update(),
	v_loc(0),
	v_parent(this),
	childs(),
	v_areas() {
	memset	(v_name, '\0', name_size + 1);
	v_flags.dir = true;
	touch();
	opendir();
	entry* idx = find(fidx);
	if(idx != nullptr) {
		v_creation	= idx->v_creation;
		v_access	= idx->v_access;
		v_update	= idx->v_update;
		unsigned char lab[slab];
		if(!idx->data(reinterpret_cast<char*>(lab), true, 0, idx->size()))
			fatx_context::get()->par.label(lab, idx->size());
	}
}
							// existing entry constructor
							entry::			entry(streamptr s, const char buf[ent_size]) :
	mux_B(),
	mux_D(),
	mux_E(),
	entbuf(),
	v_cptacc(0),
	v_writeopened(fatx_context::get()->mmi.prog != frontend::fuse),
	v_status(invalid),
	v_namesize(static_cast<uint8_t>(buf != nullptr ? buf[0] : 0)),
	v_flags(buf != nullptr ? buf[1] : '\0'),
	v_cluster(buf != nullptr ? byte_order<4>::litend(&buf[0x2C])() : 0),
	v_size(buf != nullptr ? byte_order<4>::litend(&buf[0x30])() : 0),
	v_creation(reinterpret_cast<const unsigned char *>((buf != nullptr ? &buf[0x34] : "\0\0\0\0"))),
	v_access(reinterpret_cast<const unsigned char *>((buf != nullptr ? &buf[0x38] : "\0\0\0\0"))),
	v_update(reinterpret_cast<const unsigned char *>((buf != nullptr ? &buf[0x3C] : "\0\0\0\0"))),
	v_loc(s),
	v_parent(nullptr),
	childs(),
	v_areas() {
	const string nchar = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!#$%&'()-.@[]^_`{}~ ";
	v_status = (
		// 0xFF or 0x00 on 2 firsts bytes = end of entries
		(buf == nullptr || (buf[0] == EOD && buf[1] == EOD) || (buf[0] == 0 && buf[1] == 0)) ? end : (
			(
				// invalid = invalid v_cluster (we scan something else than an entry) or invalid name
				v_cluster > fatx_context::get()->par.clus_fat) || any_of(buf + 2, buf + 2 + (v_namesize <= name_size ? v_namesize : 2), [&] (char c) { return nchar.find(c) == std::string::npos; }) ? invalid : (
				// valid entry = good name size + v_cluster not free in FAT
				(v_namesize <= name_size && ((v_size == 0 && v_cluster == 0) || (v_cluster != 0 && fatx_context::get()->fat->dskmap::read(v_cluster) != FLK))) ? valid : (
					(
						// recoverable entry = entry with v_cluster not allocated to something else
						(v_cluster == 0 && !v_flags.dir) || (
							v_cluster != 0 && fatx_context::get()->fat->dskmap::read(v_cluster) == FLK && (
							typeid(fatx_context::get()->fat) != typeid(memmap*) || (
								fatx_context::get()->fat->status(v_cluster) == memmap::deleted &&
								!fatx_context::get()->fat->getentry(v_cluster)->v_flags.dir &&
								fatx_context::get()->fat->getentry(v_cluster)->v_update.seq() < v_update.seq()
							))
						)
					) ? delwdata : (v_flags.dir ? invalid : delnodata)
				)
			)
		)
	);
	// cleaning name
	memset(v_name, '\0', name_size + 1);
	if(buf != nullptr)
		memcpy(v_name, &buf[2], (v_namesize <= name_size) ? v_namesize : name_size);
	if(v_status != end && ((v_namesize != deleted_size && v_namesize != strlen(v_name)) || strlen(v_name) == 0 || v_name[0] == sepdir[0]))
			v_status = invalid;
	if(v_status != end && v_status != invalid) {
		for(size_t i = 0; i < strlen(v_name); i++)
			v_name[i] = (v_name[i] == EOD) ? '\0' : nchar.find(v_name[i]) == std::string::npos ? '~' : v_name[i];
	}
	mux_B.name(mutex_buff + path());
	mux_D.name(mutex_data + path());
	mux_E.name(mutex_entr + path());
}
							// new entry constructor
							entry::			entry(const string &n, filesize s, const bool d) :
	mux_B(),
	mux_D(),
	mux_E(),
	entbuf(),
	v_cptacc(0),
	v_writeopened(fatx_context::get()->mmi.prog != frontend::fuse),
	v_status(invalid),
	v_namesize(0),
	v_flags(),
	v_cluster(0),
	v_size(d ? 0 : s),
	v_creation(),
	v_access(),
	v_update(),
	v_loc(0),
	v_parent(nullptr),
	childs(),
	v_areas() {
	v_areas = make_shared<vareas>(fatx_context::get()->fat->alloc(d ? 1 : clsarithm::siz2cls(s)));
	v_cluster = (!d && s == 0) ? FLK : v_areas->first();
	if(v_cluster == 0)
		v_size		= 0;
	v_namesize		= static_cast<uint8_t>((n.length() <= name_size) ? n.length() : name_size);
	memset			(v_name, '\0', name_size + 1);
	strncpy			(v_name, &n[0], v_namesize);
	mux_B.name(mutex_buff + string(v_name));
	mux_D.name(mutex_data + string(v_name));
	mux_E.name(mutex_entr + string(v_name));
	v_flags.dir		= d;
	touch();
	if(d && v_cluster != 0)
		entry(clsarithm::cls2ptr(v_cluster)).write();
}
							entry::			~entry() {
	flush(false);
	mux_D.lock();
	entbuf.reset();
	childs.clear();
	v_parent = nullptr;
	v_areas.reset();
	mux_D.unlock();
}

#ifndef NDEBUG
string						entry::			print() {
	return "-> entry " + ((status() == end) ? "*END*\n" : (boost::format("'%s' : %s, 0x%08X->0x%08X (%d) at %s\n		(%s, C=>%s A=>%s U=>%s, 0x%02X)\n")
		% ((status() == invalid) ? "*INVALID*" : path())
		% ((status() == invalid) ? "bad" : (status() == valid) ? "ok" : (status() == delwdata) ? "del" : "ko")
		% cluster()
		% ((cluster() != FLK && cluster() != EOC && cluster() <= fatx_context::get()->par.clus_fat) ? fatx_context::get()->fat->dskmap::read(cluster()) : 0)
		% size()
		% fatx_context::get()->dev.address(loc())
		% flags().print()
		% creation().print()
		% access().print()
		% update().print()
		% static_cast<unsigned int>(namesize())
	).str());
}
#endif
string						entry::			path() {
	return (
		(cluster() == fatx_context::get()->par.root_clus) ? string(sepdir) : (
		((parent() != nullptr) ? parent()->path() : (string(def_landf) + sepdir)) + name() + (flags().dir ? sepdir : "")
	));
}
void						entry::			opendir() {
	if(status() == end || status() == invalid || !flags().dir || cluster() == FLK || cluster() == EOC)
		return;
	bool marked		= false;
	bool bad		= false;
	for(clusptr clus_curr = cluster(); clus_curr != EOC && clus_curr != FLK && !(marked && !fatx_context::get()->mmi.recover); clus_curr = fatx_context::get()->fat->read(clus_curr)) {
		string buf = fatx_context::get()->dev.read(clsarithm::cls2ptr(clus_curr), fatx_context::get()->par.clus_size);
		for(size_t i = 0; i < fatx_context::get()->par.clus_size && !(marked && !fatx_context::get()->mmi.recover); i += ent_size) {
			entry* ent = new entry(clsarithm::cls2ptr(clus_curr) + i, &buf[i]);
			ent->parent(this);
			ent->mux_B.name(mutex_buff + ent->path());
			ent->mux_D.name(mutex_data + ent->path());
			ent->mux_E.name(mutex_entr + ent->path());
			if(ent->status() == end)
				marked = true;
			if(ent->status() == invalid && !marked)
				bad = true;
			if(ent->status() == end || ent->status() == invalid) {
				delete ent;
				ent = nullptr;
				continue;
			}
			if(marked && ent->status() == valid)
				ent->status(delwdata);
			if(!fatx_context::get()->mmi.recover && ent->status() != valid) {
				delete ent;
				ent = nullptr;
				continue;
			}
			#ifndef NDEBUG
				dbglog(ent->print());
			#endif
			if(ent->status() == valid) {
				mux_E.lock_shared();
				for(entry& e: childs) {
					if(e.status() == valid && ent->namesize() == e.namesize() && strncmp(ent->name(), e.name(), ent->namesize()) == 0) {
						// duplicate reference case
						if(fatx_context::get()->mmi.prog != frontend::fsck && !fatx_context::get()->mmi.recover)
							console::write((boost::format("Duplicate reference in same directory %s for entry %s.\n") % path() % ent->name()).str(), fatx_context::get()->mmi.dialog);
						ent->status(duplicate);
						e.status(duplicate);
						break;
					}
				}
				mux_E.unlock_shared();
			}
			if(ent->flags().dir) {
				bool del = false;
				// check if no circular reference
				for(entry* e = this; ; e = e->parent()) {
					if(e == nullptr || e->loc() == 0)
						break;
					if(e->status() != valid)
						del = true;
					if(ent->cluster() == e->cluster()) {
						// circular reference case
						if(!del) {
							console::write((boost::format("Circular reference for entry %s found in %s.") % ent->path() % path()).str(), fatx_context::get()->mmi.dialog);
							if(fatx_context::get()->mmi.prog == frontend::fsck) {
								console::write(" Remove it ?", fatx_context::get()->mmi.dialog);
								if(fatx_context::get()->mmi.getanswer(true)) {
									remfrdir(ent, false);
									delete ent;
									ent = nullptr;
									break;
								}
							}
							else
								console::write(" Skipping.\n", fatx_context::get()->mmi.dialog);
						}
						bad = true;
						delete ent;
						ent = nullptr;
						break;
					}
				}
				if(ent == nullptr)
					continue;
			}
			mux_E.lock();
			childs.push_back(ent);
			mux_E.unlock();
			ent = nullptr;
			mux_E.lock_shared();
			if(childs.back().flags().dir && childs.back().status() != delnodata) {
				// go one step deep
				childs.back().opendir();
			}
			mux_E.unlock_shared();
		}
		if(!marked && fatx_context::get()->fat->read(clus_curr) == EOC)
			marked = true;
	}
	if(status() == valid && !marked)
		closedir();
	if(childs.empty() && bad)
		status(delnodata);
}
void						entry::			closedir() {
	if(!flags().dir)
		return;
	bool closed = false;
	clusptr j = 0;
	for(
		clusptr i = cluster();
		!closed && i != EOC && i != FLK;
		i = fatx_context::get()->fat->read(i)
	) {
		string buf = fatx_context::get()->dev.read(clsarithm::cls2ptr(i), fatx_context::get()->par.clus_size);
		for(
			j = 0;
			j < fatx_context::get()->par.clus_size;
			j += ent_size
		) {
			if((buf[j + 0] == EOD && buf[j + 1] == EOD) || (buf[j + 0] == 0 && buf[j + 1] == 0)) {
				closed = true;
				break;
			}
		}
	}
	if(!closed && j != 0 && (j - (j >> (fatx_context::get()->par.clus_size >> ent_pow) << (fatx_context::get()->par.clus_size >> ent_pow)) != 0)) {
		console::write((boost::format("Directory %s supposed to be closed.") % path()).str(), fatx_context::get()->mmi.dialog);
		if(fatx_context::get()->mmi.prog == frontend::fsck) {
			console::write(" Close it ?", fatx_context::get()->mmi.dialog);
			if(fatx_context::get()->mmi.getanswer(true))
				entry(j).write();
		}
		else {
			console::write(" Closing.\n", fatx_context::get()->mmi.dialog);
			mux_E.lock();
			entry(j).write();
			mux_E.unlock();
		}
	}
}
int							entry::			addtodir(entry* e) {
	if(e == nullptr || !flags().dir || cluster() == 0 || (e->flags().dir && e->cluster() == 0))
		return EFAULT;
	mux_D.lock();
	mux_E.lock_shared();
	for(entry& i: childs) {
		if(i.namesize() == e->namesize() && strncmp(i.name(), e->name(), i.namesize()) == 0) {
			mux_E.unlock_shared();
			mux_D.unlock();
			return EEXIST;
		}
	}
	mux_E.unlock_shared();
	streamptr endp = 0;
	streamptr del = 0;
	for(clusptr i = cluster(); endp == 0 && i != EOC && i != FLK; i = fatx_context::get()->fat->read(i)) {
		string buf = fatx_context::get()->dev.read(clsarithm::cls2ptr(i), fatx_context::get()->par.clus_size);
		for(size_t j = 0; j < fatx_context::get()->par.clus_size; j += ent_size) {
			if(buf[j] == EOD && buf[j + 1] == EOD) {
				endp = clsarithm::cls2ptr(i) + j;
				break;
			}
			if(static_cast<unsigned char>(buf[j]) == deleted_size)
				del = clsarithm::cls2ptr(i) + j;
		}
	}
	if(endp != 0) {
		// we have enough space in directory to add one more entry
		e->loc(endp);
		e->status(entry::valid);
		if(
			(((endp + ent_size - fatx_context::get()->par.root_start) >> fatx_context::get()->par.clus_pow) << fatx_context::get()->par.clus_pow) !=
			(endp + ent_size - fatx_context::get()->par.root_start)
		) {
			// we must mark the end of entries
			mux_E.lock();
			entry(e->loc() + ent_size).write();
			mux_E.unlock();
		}
	}
	else {
		// we search a deleted entry
		if(del != 0) {
			// we found one, and use it
			e->loc(del);
			e->status(entry::valid);
		}
		else {
			// we have to allocate one cluster more in the directory
			if(!areas() || areas()->empty()) {
				areas(make_shared<vareas>(fatx_context::get()->fat->getareas(cluster())));
				if(areas()->empty()) {
					mux_D.unlock();
					return EFAULT;
				}
			}
			int res = 0;
			if((res = fatx_context::get()->fat->resize(areas(), areas()->nbcls() + 1))) {
				mux_D.unlock();
				return res;
			}
			e->loc(clsarithm::cls2ptr(areas()->last()));
			e->status(entry::valid);
			// we must mark the end of entries
			mux_E.lock();
			entry(e->loc() + ent_size).write();
			mux_E.unlock();
		}
	}
	e->parent(this);
	mux_E.lock();
	childs.push_back(e);
	e->write();
	mux_E.unlock();
	touch(false, false, true);
	int res = save();
	mux_D.unlock();
	return res;
}
void						entry::			remfrdir(entry* e, bool c) {
	if((e->status() != valid && e->status() != duplicate) || e->flags().lab)
		return;
	if(c) {
		mux_E.lock_shared();
		for(entry& f: e->childs) {
			mux_E.unlock_shared();
			e->remfrdir(&f);
			mux_E.lock_shared();
		}
		mux_E.unlock_shared();
	}
	mux_D.lock();
	if(e->cluster() != FLK)
		fatx_context::get()->fat->freefat(e->cluster());
	e->status(e->cluster() == FLK ? delnodata : delwdata);
	mux_E.lock();
	e->write();
	mux_E.unlock();
	if(c) {
		mux_E.lock_shared();
		auto i = find_if(childs.begin(), childs.end(), [e] (entry& a) -> bool { return &a == e; });
		assert(i != childs.end());
		mux_E.unlock_shared();
		mux_E.lock();
		childs.release(i).release();
		mux_E.unlock();
	}
	touch(false, false, true);
	save();
	mux_D.unlock();
	return;
}
entry*						entry::			find(const char* path) {
	entry*									res = this;
	char_separator<char>					sep(sepdir);
	string									src(path);
	if(src.rfind(sepdir, src.size()) != string::npos && src.rfind(sepdir, src.size()) == (src.size() - 1))
		src.erase(src.size() - 1);
	tokenizer<char_separator<char> >		dirs(src, sep);
	bool									found = true;
	for (const string &d : dirs) {
		found = false;
		res->mux_E.lock_shared();
		for(entry& e: res->childs) {
			if(e.status() == entry::valid && ((fatx_context::get()->mmi.cutname) ? e.name() == d.substr(0, name_size) : e.name() == d)) {
				found = true;
				res->mux_E.unlock_shared();
				res	= &e;
				break;
			}
		}
		if(!found && fatx_context::get()->mmi.recover) {
			for(entry& e: res->childs) {
				if(e.name() == d.substr(0, name_size)) {
					found = true;
					res->mux_E.unlock_shared();
					res	= &e;
					break;
				}
			}
		}
		if(!found) {
			res->mux_E.unlock_shared();
			break;
		}
	}
	if(!found)
		res = nullptr;
	return res;
}
void						entry::			touch(bool cre, bool acc, bool upd) {
	time_t t = std::time(nullptr);
	if(cre)
		creation(date(t));
	if(acc)
		access(date(t));
	if(upd)
		update(date(t));
}
int							entry::			write() {
	if(status() == invalid)
		return EFAULT;
	if(loc() == 0)
		return 0;
	string buf(ent_size, '\0');
	if(status() == end)
		memset(&buf[0], EOD, ent_size);
	else {
		touch(false, true, false);
		buf[0] = static_cast<char>((status() == delwdata || status() == delnodata) ? deleted_size : strlen(name()));
		flags().write(&buf[1]);
		memcpy(&buf[2], name(), name_size);
		memcpy(&buf[0x2C], byte_order<4>::litend(static_cast<byte_order<4>::value_type>(cluster())).data(), 4);
		memcpy(&buf[0x30], byte_order<4>::litend(static_cast<byte_order<4>::value_type>(size())).data(), 4);
		creation().write(reinterpret_cast<unsigned char*>(&buf[0x34]));
		access().write(reinterpret_cast<unsigned char*>(&buf[0x38]));
		update().write(reinterpret_cast<unsigned char*>(&buf[0x3C]));
	}
	int res = fatx_context::get()->dev.write(loc(), buf);
	return res;
}
int							entry::			save() {
	if(parent() == this)
		return 0;
	assert(parent() != nullptr);
	parent()->mux_D.lock();
	// mux_E.lock();
	int res = write();
	// mux_E.unlock();
	parent()->mux_D.unlock();
	return res;
}
int							entry::			rename(const char* n) {
	if(status() != valid)
		return EFAULT;
	string nstr = string(n);
	if(nstr.empty() || flags().lab)
		return 0;
	if(nstr.rfind(sepdir, nstr.size()) != string::npos) {
		assert(parent() != nullptr);
		entry* newpar = fatx_context::get()->root->find(nstr.substr(0, nstr.rfind(sepdir, nstr.size())).data());
		entry* oldpar = parent();
		if(newpar == nullptr)
			return ENOENT;
		if(oldpar != newpar) {
			oldpar->mux_D.lock();
			status(delwdata);
			oldpar->mux_E.lock_shared();
			auto i = find_if(oldpar->childs.begin(), oldpar->childs.end(), [this] (entry& a) -> bool { return &a == this; });
			assert(i != oldpar->childs.end());
			oldpar->mux_E.unlock_shared();
			write();
			oldpar->mux_E.lock();
			auto me = oldpar->childs.release(i);
			oldpar->mux_E.unlock();
			status(valid);
			oldpar->touch(false, false, true);
			oldpar->mux_D.unlock();
			int res = 0;
			if((res = oldpar->save()))
				return res;
			if((res = newpar->addtodir(me.get()))) {
				me.release();
				return res;
			}
			me.release();
		}
		nstr.erase(0, nstr.rfind(sepdir, nstr.size()) + 1);
	}
	if(!fatx_context::get()->mmi.cutname && nstr.size() > name_size)
		return -ENAMETOOLONG;
	name(string(name_size + 1, '\0').data());
	name(nstr.data());
	namesize(static_cast<uint8_t>((status() == valid) ? ((nstr.size() <= name_size) ? nstr.size() : name_size) : namesize()));
	return save();
}
void						entry::			recover() {
	if(fatx_context::get()->mmi.local) {
		if(!flags().dir) {
			ifstream fr;
			filesystem::create_directories(filesystem::path("./" + path()).remove_filename());
			fr.open("./" + path());
			if(fr) {
				fr.close();
				console::write("Can't open file for writing, file already exists locally.\n", true);
			}
			else {
				ofstream f("./" + path(), ios::binary | ios::trunc);
				string s(size(), '\0');
				data(&s[0], true, 0, size());
				f << s;
				f.close();
			}
		}
	}
	else {
		entry* e = parent()->find(name());
		if(e != nullptr && e->status() == entry::valid) {
			console::write("Can't restore file. Another valid file with same name exists in this directory.\n", true);
		}
		else {
			string buf = fatx_context::get()->dev.read(clsarithm::cls2ptr(clsarithm::ptr2cls(loc())), fatx_context::get()->par.clus_size);
			streamptr mark = 0;
			for(size_t i = 0; i < buf.size(); i+= ent_size) {
				if(buf[i] == EOD) {
					mark = clsarithm::cls2ptr(clsarithm::ptr2cls(loc())) + i;
					break;
				}
			}
			if(mark != 0 && loc() > mark) {
				entry no("_none", 0);
				no.loc(mark);
				no.status(delwdata);
				no.write();
				for(size_t i = mark + ent_size; i < clsarithm::cls2ptr(clsarithm::ptr2cls(loc())) + fatx_context::get()->par.clus_size; i+= ent_size)
					fatx_context::get()->dev.write(i, string(1, static_cast<char>(deleted_size)));
			}
			status(entry::valid);
			write();
			fatx_context::get()->fat->getareas(cluster(), [](clusptr c, clusptr v) -> void {
				fatx_context::get()->fat->write(c, v);
				dynamic_cast<memmap*>(fatx_context::get()->fat)->memchain.find(c)->second.status = memmap::modified;
			});
		}
	}
}
void						entry::			guess() {
	clusptr	p = cluster();
	clusptr s = 0;
	set<entry*> old;
	#if !defined NDEBUG && defined DBG_GUESS
		dbglog((boost::format("GUESS: %s 0x%08X (%d)\n") % path() % cluster() % clsarithm::siz2cls(size())).str());
	#endif
	for(
		clusptr q = cluster(), nb = flags().dir ? 1 : clsarithm::siz2cls(size());
		nb > 0;
		q = clsarithm::inccls(q)
	) {
		// is it free ?
		if(fatx_context::get()->fat->read(q) != FLK) {
			// check if first cluster is not available
			if(q == cluster()) {
				// if it is the case, give up, it's unrecoverable
				status(delnodata);
				if(fatx_context::get()->mmi.verbose)
					console::write(path() + (flags().dir ? sepdir : (boost::format(" (%d)") % size()).str()) + " not recoverable\n");
				return;
			}
			// check if it's a previous guess, if it is the case, use it
			if(fatx_context::get()->fat->getentry(q) != this) {
				// check if cluster is occupied by a file deleted older than ours
				if(
					fatx_context::get()->fat->status(q) == memmap::deleted &&
					!fatx_context::get()->fat->getentry(q)->flags().dir &&
					fatx_context::get()->mmi.deldate &&
					fatx_context::get()->fat->getentry(q)->update().seq() < update().seq()
				) {
					// we remember it and use that cluster
					old.insert(fatx_context::get()->fat->getentry(q));
				}
				else {
					// check if it belongs to a lost chain (occupied but not marked)
					if(!fatx_context::get()->fat->getentry(q)) {
						memmap::lost_t& mlost = dynamic_cast<memmap*>(fatx_context::get()->fat)->lost;
						memmap::lost_t::iterator lc = find_if(mlost.begin(), mlost.end(), [q] (const vareas& i) -> bool { return i.first() == q; });
						// if it's the beginning of a lost chain not exeeding the remaining size, use it
						if(lc != mlost.end() && lc->nbcls() <= nb) {
							// using lost chain
							mlost.erase(lc);
							// linking to lost chain
							fatx_context::get()->fat->change(p, this, q, memmap::deleted);
							// marking lost chain
							vareas va = fatx_context::get()->fat->getareas(
								q, [this](clusptr c, clusptr v) -> void {
									fatx_context::get()->fat->change(c, this, v, memmap::deleted);
								});
							// changing to last cluster of lost chain
							q = va.last();
							nb -= va.nbcls();
							p = q;
							continue;
						}
					}
					// cluster is occuped, skip it
					if(s == 0)
						s = q;
					continue;
				}
			}
		}
		// q is free
		if(s != 0) {
			#if !defined NDEBUG && defined DBG_GUESS
				dbglog((boost::format(" skip: 0x%08X-0x%08X [belongs to %s [%c]]\n")
					% s % (q - 1)
					% (fatx_context::get()->fat->getentry(s) == nullptr ? "lost chain" : fatx_context::get()->fat->getentry(s)->path())
					% (fatx_context::get()->fat->status(s) == memmap::deleted ? 'D' : 'X')
				).str());
			#endif
		}
		// we suppose that it belongs to our entry
		fatx_context::get()->fat->change(p, this, q, memmap::deleted);
		nb--;
		p = q;
	}
	fatx_context::get()->fat->change(p, this, EOC, memmap::deleted);
	#ifndef NDEBUG
	dbglog((boost::format("%s  Guessed clusters: %06d out of %06d for size %d (%s)\n")
		% ((clsarithm::siz2cls(size()) != 0 && fatx_context::get()->fat->getareas(cluster()).nbcls() != clsarithm::siz2cls(size())) ? "!!!" : "   ")
		% fatx_context::get()->fat->getareas(cluster()).nbcls()
		% (flags().dir ? 1 : clsarithm::siz2cls(size()))
		% size()
		% path()
	).str());
	#endif
	// we guess older deleted things pushed by our entry
	for(entry* i: old)
		i->guess();
}
bool						entry::			analyse(pass_t step, const string &header) {
	bool recovered	= false;
	if(step != findfile && flags().dir && status() == delnodata) {
		console::write((boost::format("Entry %s points to invalid data. Skipping.\n") % (header + name())).str(), fatx_context::get()->mmi.dialog);
		return false;
	}
	if(step == findfile && (status() == valid || status() == duplicate)) {
		// sanity checks
		if(flags().dir && cluster() == FLK) {
			console::write((boost::format("Entry %s has invalid cluster pointer.") % (header + name())).str(), fatx_context::get()->mmi.dialog);
			if(fatx_context::get()->mmi.prog == frontend::fsck) {
				console::write(" Remove it ?", fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.getanswer(true)) {
					assert(parent() != nullptr);
					parent()->remfrdir(this, false);
				}
			}
			else {
				console::write("\n", fatx_context::get()->mmi.dialog);
				status(delnodata);
			}
			return false;
		}
		bool dl = fatx_context::get()->mmi.dellost;
		fatx_context::get()->mmi.dellost = true;
		clusptr cnt	= fatx_context::get()->fat->getareas(cluster()).nbcls();
		fatx_context::get()->mmi.dellost = dl;
		if(!flags().dir && cnt != clsarithm::siz2cls(size())) {
			console::write((boost::format("%sntry %s has wrong size: declared %d, found %d.")
				% (status() == duplicate ? "Duplicate e" : "E")
				% path()
				% size()
				% (cnt * fatx_context::get()->par.clus_size)
			).str(), fatx_context::get()->mmi.dialog);
			if(fatx_context::get()->mmi.prog == frontend::fsck) {
				if(status() == duplicate) {
					console::write(" Remove it ?", fatx_context::get()->mmi.dialog);
					if(fatx_context::get()->mmi.getanswer(true)) {
						assert(parent() != nullptr);
						parent()->remfrdir(this, false);
						return false;
					}
				}
				console::write((boost::format(" Possible %s data, correct it ?") % ((cnt > clsarithm::siz2cls(size())) ? "extra" : "loss of")).str(), fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.getanswer(true)) {
					size(cnt * fatx_context::get()->par.clus_size);
					write();
				}
			}
			else
				console::write("\n", fatx_context::get()->mmi.dialog);
		}
		entry* s = nullptr;
		fatx_context::get()->fat->getareas(cluster(), [&s] (clusptr c, clusptr) -> void {
			if(s == nullptr)
				s = fatx_context::get()->fat->getentry(c);
		});
		if(s != nullptr) {
			console::write((boost::format("Conflict between %sentry %s and %sentry %s (shares same blocks).")
				% (status() == duplicate ? "duplicate " : "")
				% (header + name())
				% (s->status() == duplicate || s->status() == validupl ? "duplicate " : "")
				% (s->path().data())
			).str(), fatx_context::get()->mmi.dialog);
			if(fatx_context::get()->mmi.prog == frontend::fsck) {
				console::write((boost::format(" Remove %s (part of the other one) ?") % path().data()).str(), fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.getanswer(true)) {
					if(cluster() != s->cluster()) {
						assert(parent() != nullptr);
						cluster(FLK);
						parent()->remfrdir(this, false);
					}
					else {
						status(delnodata);
						cluster(FLK);
						write();
					}
				}
			}
			else {
				console::write("\n", fatx_context::get()->mmi.dialog);
				status(delnodata);
			}
			return false;
		}
		if(status() == duplicate) {
			auto e = find_if(parent()->childs.begin(), parent()->childs.end(), [this] (entry& i) -> bool { return namesize() == i.namesize() && strncmp(name(), i.name(), namesize()) == 0; });
			if(e == parent()->childs.end() || (e->status() != duplicate && e->status() != validupl))
				status(valid);		// problem already solved
			else if(e->status() == duplicate)
				status(validupl);	// waiting for analysis of the other entry
			else {
				// both entries are valid and different
				console::write((boost::format("Duplicate entry %s.") % (header + name())).str(), fatx_context::get()->mmi.dialog);
				if(fatx_context::get()->mmi.prog == frontend::unrm) {
					// consider it if different from its brother
					string t(name());
					t += "~";
					name(t.data());
					namesize(max<uint8_t>(static_cast<uint8_t>(namesize() + 1), name_size));
					console::write((boost::format(" Reading it as %s.\n") % name()).str(), fatx_context::get()->mmi.dialog);
				}
				else if(fatx_context::get()->mmi.prog == frontend::fsck) {
					// interactive way
					string t(name());
					t += "~";
					name(t.data());
					namesize(max<uint8_t>(static_cast<uint8_t>(namesize() + 1), name_size));
					console::write((boost::format(" Create it (as %s) ?") % name()).str(), fatx_context::get()->mmi.dialog);
					if(fatx_context::get()->mmi.getanswer(true))
						write();
					else {
						console::write(" Skipping.\n", fatx_context::get()->mmi.dialog);
						return false;
					}
				}
				else {
					// forget it
					console::write(" Skipping.\n", fatx_context::get()->mmi.dialog);
					return false;
				}
			}
		}
		// we mark valid entries
		if(fatx_context::get()->mmi.verbose)
			console::write(header + name() + (flags().dir ? sepdir : (boost::format(" (%d)") % size()).str()) + "\n");
		fatx_context::get()->mmi.dellost = true;
		areas(make_shared<vareas>(
			fatx_context::get()->fat->getareas(cluster(), [this](clusptr c, clusptr) -> void {
				fatx_context::get()->fat->change(c, this);
			})));
		fatx_context::get()->mmi.dellost = dl;
	}
	if(step == finddel && (status() == delwdata || status() == delnodata)) {
		if(fatx_context::get()->mmi.verbose) {
			console::write(header + name() + (flags().dir ? sepdir : (boost::format(" (%d)") % size()).str())
				+ " " + (status() == delwdata ? "deleted" : "not recoverable") + "\n"
			);
		}
		// we guess recoverable entries
		if(status() == delwdata)
			guess();
	}
	if(step == tryrecov && status() == delwdata && !flags().dir) {
		// we recover file sub-entries
		console::write(header + name() + (boost::format(" (%d)") % size()).str() + " recover ?");
		if(fatx_context::get()->mmi.getanswer()) {
			recovered = true;
			recover();
		}
	}
	if(flags().dir) {
		for(entry& ent: childs) {
			// we go one step deeper
			recovered = ent.analyse(step, header + name() + sepdir) || recovered;
		}
	}
	if(recovered && flags().dir && status() != valid && !fatx_context::get()->mmi.local) {
		console::write((boost::format("Recovering parent directory %s.\n") % (header + name())).str(), fatx_context::get()->mmi.dialog);
		recover();
	}
	return recovered;
}
int							entry::			resize(const filesize s) {
	#ifndef NDEBUG
		dbglog((boost::format("RESIZE: %s size:%d->%d\n") % path() % size() % s).str());
	#endif
	if(flags().dir)
		return EISDIR;
	if(!writeopened()) {
		#ifndef NDEBUG
			dbglog((boost::format("**> writing a file not opened for write (%s)\n") % path()).str());
		#endif
		return EACCES;
	}
	if(s == size())
		return 0;
	if(s == 0) {
		areas(nullptr);
		fatx_context::get()->fat->freefat(cluster());
		cluster(0);
		size(s);
	}
	else if(size() == 0) {
		vareas v = fatx_context::get()->fat->alloc(clsarithm::siz2cls(s));
		if(v.empty())
			return ENOSPC;
		cluster(v.first());
		size(s);
		areas(make_shared<vareas>(v.sub(size())));
	}
	if(s != size()) {
		int res = 0;
		if((res = fatx_context::get()->fat->resize(areas(), clsarithm::siz2cls(s))))
			return res;
		size(s);
		areas(make_shared<vareas>(areas()->sub(size())));
	}
	return save();
}
int							entry::			data(char* buf, bool r, filesize offset, filesize s) {
	s = (s == 0) ? (r ? size() - offset : size()) : (r ? min<filesize>(s, size() - offset) : s);
	#ifndef NDEBUG
		dbglog((boost::format("DATA: %s(%d) [%c:0x%08X(%d)]\n") % path() % size() % (r ? 'R' : 'W') % offset % s).str());
	#endif
	if(flags().dir)
		return EISDIR;
	if(!r && !writeopened())
		return EACCES;
	int res = 0;
	if(!r && (size() == 0 || (offset + s) > size())) {
		if((res = resize(offset + s))) {
			#ifndef NDEBUG
				dbglog((boost::format("DATA:%s resize failed.\n") % path()).str());
			#endif
			return res;
		}
	}
	if(size() != 0) {
		if(!areas() || areas()->empty()) {
			areas(make_shared<vareas>(fatx_context::get()->fat->getareas(cluster()).sub(size())));
			if(areas()->empty())
				return EFAULT;
		}
		for(const area& i: areas()->sub(s, offset)) {
			if(r)
				memcpy(buf + i.offset - offset, fatx_context::get()->dev.read(i.pointer, i.size).data(), i.size);
			else {
				if((res = fatx_context::get()->dev.write(i.pointer, string(buf + i.offset - offset, i.size))))
					return res;
			}
		}
	}
	if(!r) {
		touch(false, false, true);
		res = save();
	}
	return res;
}
size_t						entry::			bufread(char* buf, filesize offset, filesize s) {
	s = min<filesize>(size(), offset + s) - offset;
	if(offset >= size() || s == 0)
		return 0;
	mux_B.lock();
	if(entbuf && (offset < entbuf->offset || (offset + s) > (entbuf->offset + entbuf->size()))) {
		if(flush(false)) {
			#ifndef NDEBUG
				dbglog((boost::format("**> flush buffer failed (%s)") % path()).str());
			#endif
			mux_B.unlock();
			return 0;
		}
		entbuf.reset();
	}
	if(!entbuf) {
		entbuf.reset(new buffer(offset, size() - offset));
		if(entbuf->size() == 0 || data(entbuf->data(), true, entbuf->offset, entbuf->size())) {
			#ifndef NDEBUG
				dbglog((boost::format("**> alloc buffer or read operation failed (%s: 0x%08X %d)") % path() % offset % s).str());
			#endif
			mux_B.unlock();
			return 0;
		}
		#ifndef NDEBUG
			dbglog((boost::format("--> new read buffer (%s at 0x%016X: %d)\n") % path() %  entbuf.get() % entbuf->size()).str());
		#endif
	}
	s = min<filesize>(s, entbuf->size());
	memcpy(buf, &(*entbuf)[offset - entbuf->offset], s);
	#ifndef NDEBUG
		dbglog((boost::format("--> read buffer (%s at 0x%016X(%d): %d(%d))\n") % path() %  entbuf.get() % entbuf->size() % (offset - entbuf->offset) % s).str());
	#endif
	#if !defined NDEBUG && defined DBGBUFDMP
		(*entbuf)(offset - entbuf->offset);
	#endif
	mux_B.unlock();
	return s;
}
size_t						entry::			bufwrite(const char* buf, filesize offset, filesize s) {
	if(!writeopened()) {
		#ifndef NDEBUG
			dbglog((boost::format("**> writing a file not opened for write (%s)") % path()).str());
		#endif
		return false;
	}
	mux_B.lock();
	if(size() < offset + s && resize(offset + s)) {
		#ifndef NDEBUG
			dbglog((boost::format("**> file resize failed (%s: 0x%08X %d)") % path() % offset % s).str());
		#endif
		mux_B.unlock();
		return 0;
	}
	int res = 0;
	if(entbuf) {
		if(entbuf->offset + entbuf->size() == offset) {
			entbuf->enlarge(entbuf->size() + s);
			if(entbuf->offset + entbuf->size() < offset + s) {
				res = flush(false);
				entbuf.reset();
			}
		}
		else {
			res = flush(false);
			entbuf.reset();
		}
		if(res) {
			#ifndef NDEBUG
				dbglog((boost::format("**> flush buffer failed (%s)") % path()).str());
			#endif
			mux_B.unlock();
			return 0;
		}
	}
	if(!entbuf) {
		entbuf.reset(new buffer(offset, s));
		if(entbuf->size() < s) {
			#ifndef NDEBUG
				dbglog((boost::format("**> alloc buffer failed (%s: 0x%08X %d)") % path() % offset % s).str());
			#endif
			mux_B.unlock();
			return 0;
		}
	}
	s = min<filesize>(s, entbuf->size());
	memcpy(&(*entbuf)[offset - entbuf->offset], buf, s);
	entbuf->touched = true;
	#ifndef NDEBUG
		dbglog((boost::format("--> write buffer (%s at 0x%08X(%d): %d(%d))\n") % path() %  entbuf.get() % entbuf->size() % (offset - entbuf->offset) % s).str());
	#endif
	#if !defined NDEBUG && defined DBGBUFDMP
		(*entbuf)(offset - entbuf->offset);
	#endif
	mux_B.unlock();
	return s;
}
int							entry::			flush(bool l) {
	if(flags().dir)
		return 0;
	int res = 0;
	if(l)
		mux_B.lock();
	if(entbuf) {
		#ifndef NDEBUG
		if(entbuf->touched)
			dbglog((boost::format("<-> flush buffer (%s at 0x%016X: %d)\n") % path() % entbuf.get() % entbuf->size()).str());
		else
			dbglog((boost::format("<-> no need to flush buffer (%s at 0x%016X: %d)\n") % path() % entbuf.get() % entbuf->size()).str());
		#endif
		if(entbuf->touched) {
			if(writeopened()) {
				res = data(entbuf.get()->data(), false, entbuf->offset, entbuf->size());
				if(!res)
					entbuf.reset();
			}
			else
				res = EACCES;
		}
	}
	if(l)
		mux_B.unlock();
	return res;
}
void						entry::			open(bool w) {
	if(!flags().dir) {
		if(w)
			mux_D.lock();
		else
			mux_D.lock_shared();
		if(!writeopened())
			writeopened(w);
		if(cptacc() == 0 && cluster() != 0 && size() != 0)
			areas(make_shared<vareas>(fatx_context::get()->fat->getareas(cluster()).sub(size())));
		cptacc(cptacc() + 1);
	}
}
void						entry::			close(bool w) {
	if(!flags().dir) {
		if(writeopened()) {
			flush(true);
			writeopened(false);
		}
		cptacc(cptacc() - 1);
		if(cptacc() == 0) {
			areas(nullptr);
			if(entbuf) {
				mux_B.lock();
				entbuf.reset();
				mux_B.unlock();
			}
		}
		if(w)
			mux_D.unlock();
		else
			mux_D.unlock_shared();
	}
}
bool						entry::			operator == (entry& b) {
	bool res =
		loc() == b.loc() &&
		cluster() == b.cluster() &&
		size() == b.size() &&
		parent() == b.parent() &&
		strncmp(name(), b.name(), namesize()) == 0
	;
	return res;
}

static int									fatx_getattr	(const char* path, struct stat* st) {
	#ifndef NDEBUG
		dbglog((boost::format("GETATTR: %s\n") % path).str());
	#endif
	entry* f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid || (f->flags().dir && f->cluster() == 0))
		return -ENOENT;
	memset(st, 0, sizeof(struct stat));
	st->st_dev		= fatx_context::get()->par.par_id;
	st->st_mode		= f->flags()() & (fatx_context::get()->mmi.mask | S_IFDIR | S_IFREG);
	st->st_nlink	= f->childs.size() + 1;
	st->st_size		= static_cast<__off_t>(f->flags().dir ? f->childs.size() : f->size());
	st->st_blksize	= fatx_context::get()->par.clus_size;
	st->st_blocks	= static_cast<__blkcnt_t>(clsarithm::siz2cls(f->size()) * fatx_context::get()->par.clus_size / blksize);
	st->st_atime	= f->access()();
	st->st_mtime	= f->update()();
	st->st_ctime	= f->creation()();
	st->st_uid		= fatx_context::get()->mmi.uid;
	st->st_gid		= fatx_context::get()->mmi.gid;
	#ifndef NDEBUG
		dbglog(f->print());
	#endif
	return 0;
}
static int									fatx_truncate	(const char* path, off_t size) {
	#ifndef NDEBUG
		dbglog((boost::format("TRUNCATE: %s\n") % path).str());
	#endif
	entry* f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	if(!fatx_context::get()->mmi.writeable())
		return -EROFS;
	if(f->flags().ro)
		return -EACCES;
	return -f->resize(static_cast<filesize>(size));
}
static int									fatx_open		(const char* path, struct fuse_file_info* fi) {
	#ifndef NDEBUG
		dbglog((boost::format("OPEN: %s [%c]\n") % path % (fi->flags & (S_IWUSR | S_IWGRP | S_IWOTH) ? 'w' : 'r' )).str());
	#endif
	entry* f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	if((fi->flags & (O_WRONLY | O_RDWR)) != 0 && !fatx_context::get()->mmi.writeable())
		return -EROFS;
	if((fi->flags & (O_WRONLY | O_RDWR)) != 0 && f->flags().ro)
		return -EPERM;
	f->open((fi->flags & (O_WRONLY | O_RDWR)) != 0);
	fi->fh = uint64_t(f);
	return 0;
}
static int									fatx_flush		(const char* path, struct fuse_file_info* fi) {
	#ifndef NDEBUG
		dbglog((boost::format("FLUSH: %s\n") % path).str());
	#endif
	entry* f(reinterpret_cast<entry *>(fi->fh));
	if(f == nullptr)
		f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	return -f->flush();
}
static int									fatx_close		(const char* path, struct fuse_file_info* fi) {
	#ifndef NDEBUG
		dbglog((boost::format("CLOSE: %s\n") % path).str());
	#endif
	entry* f(reinterpret_cast<entry *>(fi->fh));
	if(f == nullptr)
		f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	f->close((fi->flags & (O_WRONLY | O_RDWR)) != 0);
	fi->fh = uint64_t(0);
	return 0;
}
static int									fatx_read		(const char* path, char* buf, size_t size, off_t offset, struct fuse_file_info* fi) {
	#ifndef NDEBUG
		dbglog((boost::format("READ: %s (%d@%d)\n") % path % size % offset).str());
	#endif
	entry* f(reinterpret_cast<entry *>(fi->fh));
	if(f == nullptr)
		f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	return static_cast<int>(f->bufread(buf, static_cast<filesize>(offset), size));
}
static int									fatx_write		(const char* path, const char* buf, size_t size, off_t offset, struct fuse_file_info* fi) {
	#ifndef NDEBUG
		dbglog((boost::format("WRITE: %s (%d@%d)\n") % path % size % offset).str());
	#endif
	entry* f(reinterpret_cast<entry*>(fi->fh));
	if(f == nullptr)
		f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	if(!fatx_context::get()->mmi.writeable())
		return -EROFS;
	if(f->flags().ro)
		return -EACCES;
	return static_cast<int>(f->bufwrite(buf, static_cast<filesize>(offset), size));
}
static int									fatx_chmod		(const char* path, mode_t mode) {
	#ifndef NDEBUG
		dbglog((boost::format("CHMOD: %s\n") % path).str());
	#endif
	entry* f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	if(!fatx_context::get()->mmi.writeable())
		return -EROFS;
	f->flags()(mode);
	return -f->save();
}
static int									fatx_chown		(const char* path, uid_t, gid_t) {
	#ifndef NDEBUG
		dbglog((boost::format("CHOWN: %s\n") % path).str());
	#endif
	entry* f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	if(!fatx_context::get()->mmi.writeable())
		return -EROFS;
	// nothing done
	return 0;
}
static int									fatx_readdir	(const char* path, void* buf, fuse_fill_dir_t ff, off_t, struct fuse_file_info* fi) {
	#ifndef NDEBUG
		dbglog((boost::format("READDIR: %s\n") % path).str());
	#endif
	entry* f(reinterpret_cast<entry*>(fi->fh));
	if(f == nullptr)
		f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	struct stat st;
	int res;
	if((res = fatx_getattr(f->path().data(), &st)) != 0)
		return res;
	ff(buf, ".", &st, 0);
	if((res = fatx_getattr(f->parent()->path().data(), &st)) != 0)
		return res;
	ff(buf, "..", &st, 0);
	for(entry& i: f->childs) {
		if(i.status() == entry::valid || (fatx_context::get()->mmi.recover && i.status() == entry::delwdata)) {
			if((res = fatx_getattr(i.path().data(), &st)) != 0)
				return res;
			if(ff(buf, i.name(), &st, 0) != 0)
				return -EBADF;
			#ifndef NDEBUG
				dbglog((boost::format(" %s\n") % i.path()).str());
			#endif
		}
	}
	return 0;
}
static int									fatx_create		(const char* path, mode_t mode) {
	#ifndef NDEBUG
		dbglog((boost::format("CREATE: %s\n") % path).str());
	#endif
	if(!fatx_context::get()->mmi.writeable())
		return -EACCES;
	string p(path);
	size_t l = p.find_last_of(sepdir);
	if(l == string::npos || l == p.length() - 1)
		return -ENOENT;
	if(fatx_context::get()->mmi.cutname)
		p.resize(l + 1 + name_size);
	else {
		if((p.length() - (l + 1)) > name_size)
			return -ENAMETOOLONG;
	}
	if(fatx_context::get()->root->find(path) != nullptr)
		return -EEXIST;
	entry* n = new entry(p.substr(l + 1), 0, ((mode & S_IFREG) == 0));
	if(n->flags().dir && n->cluster() == 0) {
		delete n;
		return -ENOSPC;
	}
	entry* s = fatx_context::get()->root->find(p.substr(0, l).data());
	if(s == nullptr) {
		delete n;
		return -ENOENT;
	}
	int res = 0;
	if((res = s->addtodir(n))) {
		delete n;
		return -res;
	}
	return fatx_chmod(path, mode);
}
static int									fatx_creope		(const char* path, mode_t mode, struct fuse_file_info* fi) {
	#ifndef NDEBUG
		dbglog((boost::format("CREOPE: %s\n") % path).str());
	#endif
	int res = 0;
	if((res = fatx_create(path, mode)) != 0)
		return res;
	return fatx_open(path, fi);
}
static int									fatx_remove		(const char* path) {
	#ifndef NDEBUG
		dbglog((boost::format("REMOVE: %s\n") % path).str());
	#endif
	entry* f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	if(!fatx_context::get()->mmi.writeable())
		return -EROFS;
	if(f->flags().ro)
		return -EACCES;
	if(f == fatx_context::get()->root)
		return -EBUSY;
	if(f->flags().dir && f->childs.size() != 0)
		return -ENOTEMPTY;
	assert(f->parent() != nullptr);
	f->parent()->remfrdir(f);
	return 0;
}
static int									fatx_rename		(const char* from, const char* to) {
	#ifndef NDEBUG
		dbglog((boost::format("RENAME: %s to %s\n") % from % to).str());
	#endif
	entry* f = fatx_context::get()->root->find(from);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	if(!fatx_context::get()->mmi.writeable())
		return -EROFS;
	if(f->flags().ro)
		return -EACCES;
	return -f->rename(to);
}
static int									fatx_utimens	(const char* path, const struct timespec tv[2]) {
	#ifndef NDEBUG
		dbglog((boost::format("UTIMENS: %s\n") % path).str());
	#endif
	entry* f = fatx_context::get()->root->find(path);
	if(f == nullptr || f->status() == entry::invalid)
		return -ENOENT;
	if(!fatx_context::get()->mmi.writeable())
		return -EROFS;
	if(f->flags().ro)
		return -EACCES;
	date d;
	d(tv[0].tv_sec);
	f->access(d);
	d(tv[1].tv_sec);
	f->update(d);
	return -f->save();
}
static int									fatx_statfs		(const char* path, struct statvfs* sfs) {
	#ifndef NDEBUG
		dbglog((boost::format("STATFS: %s\n") % path).str());
	#else
		(void)path;
	#endif
	sfs->f_bsize	= 1;									// blksize;
	sfs->f_frsize	= 1;									// fatx_context::get()->par.clus_size / sfs->f_bsize;
	sfs->f_blocks	= (fatx_context::get()->par.clus_fat - fatx_context::get()->par.root_clus) * fatx_context::get()->par.clus_size;
	sfs->f_bfree	= fatx_context::get()->fat->clsavail() * fatx_context::get()->par.clus_size;
	sfs->f_bavail	= fatx_context::get()->fat->clsavail() * fatx_context::get()->par.clus_size;
	sfs->f_files	= 0;
	sfs->f_ffree	= 0;
	sfs->f_favail	= 0;
	sfs->f_fsid		= fatx_context::get()->par.par_id;
	sfs->f_flag		= ST_NOSUID | (fatx_context::get()->mmi.writeable() ? 0 : ST_RDONLY);
	sfs->f_namemax	= name_size;
	#ifndef NDEBUG
		dbglog((boost::format(
			" f_bsize:\t%d\n"
			" f_frsize:\t%d\n"
			" f_blocks:\t%d\n"
			" f_bfree:\t%d\n"
			" f_bavail:\t%d\n"
			" f_files:\t%d\n"
			" f_ffree:\t%d\n"
			" f_favail:\t%d\n"
			" f_fsid:\t0x%08X\n"
			" f_flag:\t0x%08X\n"
			" f_namemax:\t%d\n"
		)
			% sfs->f_bsize
			% sfs->f_frsize
			% sfs->f_blocks
			% sfs->f_bfree
			% sfs->f_bavail
			% sfs->f_files
			% sfs->f_ffree
			% sfs->f_favail
			% sfs->f_fsid
			% sfs->f_flag
			% sfs->f_namemax
		).str());
	#endif
	return 0;
}
static void*								fatx_init		(struct fuse_conn_info* fci) {
	#ifndef NDEBUG
		dbglog("INIT\n");
	#endif
	fci->want = FUSE_CAP_BIG_WRITES | FUSE_CAP_DONT_MASK;
	return nullptr;
}
static void									fatx_destroy	(void*) {
	#ifndef NDEBUG
		dbglog("DESTROY\n");
	#endif
	fatx_context::get()->destroy();
	delete fatx_context::get();
}

static struct fuse_operations				fatx_ops;

int											main(int argc, char* argv[]) {
	int err = 0;
	frontend mmi(argc, argv);
	if((err = mmi.setup()))
		return err;
	if(mmi.prog != frontend::unknown && mmi.prog != frontend::label)
		console::write("Analysing filesystem, please wait.\n");
	if(mmi.prog != frontend::unknown) {
		fatx_context::set(new fatx_context(mmi));
		if((err = fatx_context::get()->setup())) {
			delete fatx_context::get();
			return err;
		}
	}
	if(mmi.prog == frontend::fsck || (mmi.prog != frontend::unknown && mmi.recover)) {
		console::write("Finding all files and directories.\n");
		fatx_context::get()->root->analyse(entry::findfile);
		#if !defined NDEBUG && defined DBG_FAT
			fatx_context::get()->fat->printfat();
		#endif
	}
	if(mmi.prog != frontend::unknown && mmi.recover) {
		console::write("Finding all deleted files and directories.\n");
		fatx_context::get()->fat->fatlost();
		fatx_context::get()->root->analyse(entry::finddel);
		#if !defined NDEBUG && defined DBG_FAT
			fatx_context::get()->fat->printfat();
		#endif
	}
	if(mmi.prog != frontend::mkfs && !mmi.script.empty()) {
		fatx_context::get()->fat->gapcheck();
		mmi.parser();
	}
	if(mmi.prog == frontend::unrm) {
		console::write("Trying to recover deleted files and directories.\n");
		if(!mmi.local)
			fatx_context::get()->fat->gapcheck();
		fatx_context::get()->root->analyse(entry::tryrecov);
	}
	if(!mmi.nofat && (mmi.prog == frontend::fsck || mmi.prog == frontend::unrm)) {
		console::write("Checking FAT consistency.\n");
		fatx_context::get()->fat->fatlost();
		fatx_context::get()->fat->fatcheck();
	}
	if(mmi.prog == frontend::fuse) {
		int fuse_argc = 0;
		char* fuse_argv[max_fuse_args];
		vector<unique_ptr<string>> vp;
		fuse_argv[fuse_argc++] = &mmi.progname[0];
		fuse_argv[fuse_argc++] = &mmi.mount[0];
		if(!mmi.writeable()) {
			vp.push_back(make_unique<string>("-o"));
			fuse_argv[fuse_argc++] = vp.back().get()->data();
			vp.push_back(make_unique<string>("ro"));
			fuse_argv[fuse_argc++] = vp.back().get()->data();
		}
		if(mmi.fuse_debug) {
			vp.push_back(make_unique<string>("-d"));
			fuse_argv[fuse_argc++] = vp.back().get()->data();
		}
		if(mmi.fuse_foregrd) {
			vp.push_back(make_unique<string>("-f"));
			fuse_argv[fuse_argc++] = vp.back().get()->data();
		}
		if(mmi.fuse_singlethr) {
			vp.push_back(make_unique<string>("-s"));
			fuse_argv[fuse_argc++] = vp.back().get()->data();
		}
		if(!mmi.fuse_option.empty()) {
			vp.push_back(make_unique<string>("-o"));
			fuse_argv[fuse_argc++] = vp.back().get()->data();
			fuse_argv[fuse_argc++] = &mmi.fuse_option[0];
		}
		for(const string& i: mmi.unkopt) {
			if(fuse_argc < (max_fuse_args - 1)) {
				vp.push_back(make_unique<string>(i));
				fuse_argv[fuse_argc++] = vp.back().get()->data();
			}
			else {
				console::write("Too many arguments for fuse. Ignoring last arguments.\n", true);
				break;
			}
		}
		fuse_argv[fuse_argc] = nullptr;
		if(!mmi.recover)
			fatx_context::get()->fat->gapcheck();
		memset(&fatx_ops, 0, sizeof(fatx_ops));
		fatx_ops.getattr		= fatx_getattr;
		fatx_ops.utimens		= fatx_utimens;
		fatx_ops.chmod			= fatx_chmod;
		fatx_ops.chown			= fatx_chown;
		fatx_ops.create			= fatx_creope;
		fatx_ops.open			= fatx_open;
		fatx_ops.read			= fatx_read;
		fatx_ops.write			= fatx_write;
		fatx_ops.flush			= fatx_flush;
		fatx_ops.release		= fatx_close;
		fatx_ops.truncate		= fatx_truncate;
		fatx_ops.unlink			= fatx_remove;
		fatx_ops.mkdir			= fatx_create;
		fatx_ops.opendir		= fatx_open;
		fatx_ops.readdir		= fatx_readdir;
		fatx_ops.releasedir		= fatx_close;
		fatx_ops.rmdir			= fatx_remove;
		fatx_ops.rename			= fatx_rename;
		fatx_ops.statfs			= fatx_statfs;
		fatx_ops.init			= fatx_init;
		fatx_ops.destroy		= fatx_destroy;
		#ifndef NDEBUG
			string s;
			for(int i = 0; i < fuse_argc; i++)
				s += string(" ") + fuse_argv[i];
			dbglog((boost::format("Fuse called with%s\n:") % s).str());
		#endif
		console::write("Ready.\n");
		fatx_context::get()->ready = true;

		err = fuse_main(fuse_argc, fuse_argv, &fatx_ops, nullptr);
		#ifndef NDEBUG
			dbglog((boost::format("Fuse returned: %d\n") % err).str());
		#endif
	}
	bool answ = false;
	if(mmi.prog == frontend::mkfs) {
		console::write((boost::format("Are you sure you want to erase all data in %s ?") % mmi.input).str());
		if((answ = mmi.getanswer(false))) {
			console::write("Creating new FATX filesystem");
			fatx_context::get()->par.write();
			console::write(".");
			fatx_context::get()->fat->erase();
			console::write(".");
			fatx_context::get()->dev.write(fatx_context::get()->par.root_start, string(fatx_context::get()->par.clus_size, '\0'));
			console::write(".");
			fatx_context::get()->root = new entry("", 0, true);
			fatx_context::get()->root->parent(fatx_context::get()->root);
			fatx_context::get()->root->status(entry::valid);
			console::write("done.\n");
			console::write((boost::format("FATX filesystem created with %d clusters.\n") % fatx_context::get()->par.clus_fat).str());
			if(mmi.volname.empty())
				mmi.volname = def_label;
		}
	}
	if((mmi.prog == frontend::mkfs && answ) || (mmi.prog == frontend::label && !mmi.volname.empty())) {
		if(mmi.prog == frontend::label)
			fatx_context::get()->fat->gapcheck();
		fatx_context::get()->par.par_label = fatx_context::get()->mmi.volname;
		unsigned char lab[slab];
		entry* idx = fatx_context::get()->root->find(fidx);
		filesize s = fatx_context::get()->par.label(lab);
		if(idx == nullptr) {
			fatx_context::get()->root->addtodir(new entry(fidx, 0, false));
			idx = fatx_context::get()->root->find(fidx);
			assert(idx != nullptr);
			attrib fl = idx->flags();
			fl.lab = true;
			idx->flags(fl);
			idx->save();
		}
		if(!idx->resize(s) && !idx->data(reinterpret_cast<char*>(lab), false, 0, s))
			console::write((boost::format("Volume label has been changed to %s\n") % fatx_context::get()->par.par_label).str());
		else
			console::write("Unable to change volume label.\n");
	}
	if(mmi.prog == frontend::fsck) {
		if(fatx_context::get()->par.par_label.empty())
			console::write("Warning: volume has no label.\n");
		if(mmi.verbose) {
			console::write((boost::format(
					"Volume label:\t%s\n"
					"Clusters size:\t%d\n"
					"Total clusters:\t%d\n"
					"Clusters free:\t%d\n"
				)
				% (fatx_context::get()->par.par_label.empty() ? "none" : fatx_context::get()->par.par_label)
				% fatx_context::get()->par.clus_size
				% fatx_context::get()->par.clus_fat
				% fatx_context::get()->fat->clsavail()
			).str());
		}
	}
	if(mmi.prog == frontend::label && mmi.volname.empty())
		console::write((fatx_context::get()->par.par_label.empty() ? "No volume label." : fatx_context::get()->par.par_label) + "\n");
	if(mmi.verbose) {
		if(fatx_context::get()->dev.modified())
			console::write("Changes have been made.\n");
		else
			console::write("No change has been made.\n");
	}
	if(err != 0)
		err = mmi.prog != frontend::fsck ? code_noerr : !fatx_context::get()->dev.modified() ? code_noerr : fatx_context::get()->mmi.allyes ? code_corrd : code_ncorr;
	delete fatx_context::get();
	#ifndef NDEBUG
		dbglog((boost::format("<= ENDC: %s\n") % mmi.name()).str());
	#endif
	return err;
}
